"""Minimal stdio MCP server exposing worai tools."""

from __future__ import annotations

import json
import os
import subprocess
import sys
import threading
import time
import uuid
from dataclasses import dataclass, field
import argparse
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import Any

from typer.testing import CliRunner

from worai.config import resolve_config_path

_JSONRPC = "2.0"
_MAX_OUTPUT_CHARS_DEFAULT = 20000
_MAX_OUTPUT_CHARS_LIMIT = 200000
_SYNC_TIMEOUT_SEC = 120.0
_DEFAULT_EVENT_PREFIX = "WORAI_MCP_EVENT "
_EVENT_PREFIX_ENV = "WORAI_MCP_EVENT_PREFIX"
_FORBIDDEN_ARGV_PREFIXES: tuple[tuple[str, ...], ...] = (
    ("agent", "mcp", "serve"),
    ("self", "update", "--yes"),
)


@dataclass(frozen=True)
class SessionDefaults:
    config_path: str | None = None
    profile: str | None = None


@dataclass(frozen=True)
class ToolPolicy:
    mode: str  # sync | async_required


@dataclass
class Job:
    id: str
    command: list[str]
    effective_config_path: str
    effective_profile: str
    state: str = "queued"  # queued|running|done|failed|canceled
    created_at: float = field(default_factory=time.time)
    started_at: float | None = None
    finished_at: float | None = None
    exit_code: int | None = None
    output: str = ""
    events: list[dict[str, Any]] = field(default_factory=list)
    next_event_seq: int = 1
    progress: dict[str, Any] = field(default_factory=dict)
    process: subprocess.Popen[str] | None = None
    lock: threading.Lock = field(default_factory=threading.Lock)


class McpTransport:
    """JSON-RPC over stdio transport for MCP."""

    _line_delimited_mode: bool | None = None

    @staticmethod
    def read_message() -> dict[str, Any] | None:
        first_line = sys.stdin.buffer.readline()
        if not first_line:
            return None

        stripped = first_line.strip()
        if stripped.startswith((b"{", b"[")):
            McpTransport._line_delimited_mode = True
            data = json.loads(stripped.decode("utf-8"))
            if not isinstance(data, dict):
                return None
            return data

        headers: dict[str, str] = {}
        line = first_line
        while True:
            if line in {b"\r\n", b"\n"}:
                break
            decoded = line.decode("utf-8", errors="replace").strip()
            if ":" in decoded:
                key, value = decoded.split(":", 1)
                headers[key.strip().lower()] = value.strip()
            line = sys.stdin.buffer.readline()
            if not line:
                return None

        McpTransport._line_delimited_mode = False
        content_length_raw = headers.get("content-length")
        if not content_length_raw:
            return None
        content_length = int(content_length_raw)
        payload = sys.stdin.buffer.read(content_length)
        if not payload:
            return None
        data = json.loads(payload.decode("utf-8"))
        if not isinstance(data, dict):
            return None
        return data

    @staticmethod
    def write_message(message: dict[str, Any]) -> None:
        encoded = json.dumps(message).encode("utf-8")
        if McpTransport._line_delimited_mode is True:
            sys.stdout.buffer.write(encoded + b"\n")
        else:
            header = f"Content-Length: {len(encoded)}\r\n\r\n".encode("ascii")
            sys.stdout.buffer.write(header)
            sys.stdout.buffer.write(encoded)
        sys.stdout.buffer.flush()


class _PatchedEnvContext:
    def __init__(self, updates: dict[str, str]) -> None:
        self._updates = updates
        self._original: dict[str, str | None] = {}

    def __enter__(self) -> None:
        self._original = {key: os.environ.get(key) for key in self._updates}
        os.environ.update(self._updates)

    def __exit__(self, *_args: object) -> None:
        for key, value in self._original.items():
            if value is None:
                os.environ.pop(key, None)
            else:
                os.environ[key] = value


class WoraiExecutor:
    """Executes worai commands (sync in-process, async as subprocess command)."""

    def __init__(self, defaults: SessionDefaults, runner: CliRunner) -> None:
        self._defaults = defaults
        self._runner = runner

    @staticmethod
    def _root_app():
        from worai import cli as root_cli

        return root_cli.app

    def resolve_context(
        self,
        *,
        config_path: str | None,
        profile: str | None,
    ) -> tuple[str | None, str, str, list[str]]:
        effective_input_config = config_path or self._defaults.config_path
        effective_profile = profile or self._defaults.profile or os.environ.get("WORAI_PROFILE") or "default"
        resolved_config = str(resolve_config_path(effective_input_config, env=os.environ))

        root_argv: list[str] = []
        if effective_input_config:
            root_argv.extend(["--config", effective_input_config])
        if effective_profile:
            root_argv.extend(["--profile", effective_profile])
        return effective_input_config, effective_profile, resolved_config, root_argv

    def invoke_sync(
        self,
        *,
        argv: list[str],
        config_path: str | None,
        profile: str | None,
        max_output_chars: int,
    ) -> dict[str, Any]:
        _, effective_profile, resolved_config, root_argv = self.resolve_context(
            config_path=config_path,
            profile=profile,
        )
        full_argv = [*root_argv, *argv]
        with _PatchedEnvContext(
            {
                "WORAI_DISABLE_UPDATE_CHECK": "1",
                "WORAI_NO_PROGRESS": "1",
                "WORAI_PROGRESS": "0",
            }
        ):
            result = self._runner.invoke(self._root_app(), full_argv)
        output, truncated = _truncate(result.output, max_output_chars)
        return {
            "ok": result.exit_code == 0,
            "exit_code": int(result.exit_code),
            "command": ["worai", *full_argv],
            "effective_config_path": resolved_config,
            "effective_profile": effective_profile,
            "output": output,
            "output_truncated": truncated,
        }

    def build_subprocess_command(
        self,
        *,
        argv: list[str],
        config_path: str | None,
        profile: str | None,
    ) -> tuple[list[str], str, str]:
        _, effective_profile, resolved_config, root_argv = self.resolve_context(
            config_path=config_path,
            profile=profile,
        )
        cmd = [sys.executable, "-m", "worai.cli", *root_argv, *argv]
        return cmd, resolved_config, effective_profile


class WoraiJobManager:
    """Background job manager for async-required tools."""

    def __init__(self) -> None:
        self._jobs: dict[str, Job] = {}
        self._lock = threading.Lock()

    def start(self, *, command: list[str], effective_config_path: str, effective_profile: str) -> dict[str, Any]:
        job = Job(
            id=str(uuid.uuid4()),
            command=command,
            effective_config_path=effective_config_path,
            effective_profile=effective_profile,
        )
        with self._lock:
            self._jobs[job.id] = job

        thread = threading.Thread(target=self._run, args=(job.id,), daemon=True)
        thread.start()
        return {
            "job_id": job.id,
            "state": job.state,
            "effective_config_path": effective_config_path,
            "effective_profile": effective_profile,
            "command": command,
        }

    def _run(self, job_id: str) -> None:
        job = self._jobs[job_id]
        job.state = "running"
        job.started_at = time.time()
        self._append_event(job, "job.started", {"state": "running"})

        env = os.environ.copy()
        env["WORAI_DISABLE_UPDATE_CHECK"] = "1"
        env["WORAI_NO_PROGRESS"] = "1"
        env["WORAI_PROGRESS"] = "0"
        env["WORAI_MCP_PROGRESS"] = "1"
        env.setdefault(_EVENT_PREFIX_ENV, _DEFAULT_EVENT_PREFIX)

        proc = subprocess.Popen(
            job.command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            bufsize=1,
            env=env,
        )
        job.process = proc

        assert proc.stdout is not None
        for line in proc.stdout:
            if self._consume_progress_event(job, line):
                continue
            with job.lock:
                job.output += line

        proc.wait()
        job.exit_code = int(proc.returncode)
        if job.state != "canceled":
            job.state = "done" if proc.returncode == 0 else "failed"
        self._append_event(job, "job.finished", {"state": job.state, "exit_code": job.exit_code})
        job.finished_at = time.time()

    def status(self, job_id: str) -> dict[str, Any]:
        job = self._get(job_id)
        with job.lock:
            progress = dict(job.progress)
            event_cursor = job.next_event_seq - 1
        return {
            "job_id": job.id,
            "state": job.state,
            "exit_code": job.exit_code,
            "created_at": job.created_at,
            "started_at": job.started_at,
            "finished_at": job.finished_at,
            "effective_config_path": job.effective_config_path,
            "effective_profile": job.effective_profile,
            "command": job.command,
            "progress": progress,
            "event_cursor": event_cursor,
        }

    def events(self, job_id: str, *, cursor: int = 0, limit: int = 100) -> dict[str, Any]:
        job = self._get(job_id)
        safe_cursor = max(0, int(cursor))
        safe_limit = max(1, min(int(limit), 500))
        with job.lock:
            events = [event for event in job.events if int(event.get("seq", 0)) > safe_cursor][:safe_limit]
            next_cursor = safe_cursor if not events else int(events[-1]["seq"])
        return {
            "job_id": job.id,
            "cursor": safe_cursor,
            "next_cursor": next_cursor,
            "events": events,
            "state": job.state,
        }

    def output(self, job_id: str, offset: int = 0, max_chars: int = _MAX_OUTPUT_CHARS_DEFAULT) -> dict[str, Any]:
        job = self._get(job_id)
        with job.lock:
            text = job.output
        safe_offset = max(0, min(offset, len(text)))
        chunk, truncated = _truncate(text[safe_offset:], max_chars)
        return {
            "job_id": job.id,
            "offset": safe_offset,
            "next_offset": safe_offset + len(chunk),
            "text": chunk,
            "truncated": truncated,
            "state": job.state,
        }

    def cancel(self, job_id: str) -> dict[str, Any]:
        job = self._get(job_id)
        if job.state in {"done", "failed", "canceled"}:
            return {"job_id": job.id, "state": job.state}
        proc = job.process
        if proc is not None:
            proc.terminate()
        job.state = "canceled"
        job.finished_at = time.time()
        self._append_event(job, "job.canceled", {"state": "canceled"})
        return {"job_id": job.id, "state": job.state}

    def _append_event(self, job: Job, event_name: str, meta: dict[str, Any]) -> None:
        now = time.time()
        with job.lock:
            event = {
                "seq": job.next_event_seq,
                "timestamp": now,
                "event": event_name,
                "meta": meta,
            }
            job.next_event_seq += 1
            job.events.append(event)
            if len(job.events) > 10000:
                job.events = job.events[-10000:]
            progress = self._progress_from_event(event_name, meta, now, dict(job.progress))
            if progress is not None:
                job.progress = progress

    def _consume_progress_event(self, job: Job, line: str) -> bool:
        prefix = os.environ.get(_EVENT_PREFIX_ENV, _DEFAULT_EVENT_PREFIX)
        if not line.startswith(prefix):
            return False
        raw = line[len(prefix):].strip()
        if not raw:
            return True
        try:
            payload = json.loads(raw)
        except json.JSONDecodeError:
            return True
        if not isinstance(payload, dict):
            return True
        event_name = str(payload.get("event", "")).strip()
        if not event_name:
            return True
        meta = payload.get("meta")
        safe_meta = meta if isinstance(meta, dict) else {}
        self._append_event(job, event_name, safe_meta)
        return True

    @staticmethod
    def _progress_from_event(
        event_name: str,
        meta: dict[str, Any],
        timestamp: float,
        current: dict[str, Any],
    ) -> dict[str, Any] | None:
        if event_name == "inventory.progress.started":
            total = int(meta.get("total", 0) or 0)
            return {
                "phase": "running",
                "completed": 0,
                "total": total,
                "percent": 0.0 if total > 0 else None,
                "last_update_at": timestamp,
                "last_event": event_name,
            }
        if event_name == "inventory.progress.updated":
            completed = int(meta.get("completed", 0) or 0)
            total = int(meta.get("total", completed) or completed)
            percent = None if total <= 0 else round((completed / total) * 100.0, 2)
            return {
                "phase": "running",
                "completed": completed,
                "total": total,
                "percent": percent,
                "last_update_at": timestamp,
                "last_event": event_name,
            }
        if event_name == "inventory.progress.completed":
            completed = int(meta.get("completed", current.get("completed", 0)) or 0)
            total = int(meta.get("total", current.get("total", completed)) or completed)
            percent = 100.0 if total > 0 else None
            return {
                "phase": "completed",
                "completed": completed,
                "total": total,
                "percent": percent,
                "last_update_at": timestamp,
                "last_event": event_name,
            }
        if event_name.startswith("job."):
            updated = dict(current)
            updated["last_update_at"] = timestamp
            updated["last_event"] = event_name
            return updated
        return None

    def _get(self, job_id: str) -> Job:
        with self._lock:
            job = self._jobs.get(job_id)
        if job is None:
            raise ValueError(f"Unknown job_id: {job_id}")
        return job


class WoraiMcpServer:
    """MCP server with typed tool dispatch and async jobs."""

    TOOL_POLICIES: dict[str, ToolPolicy] = {
        "worai_version": ToolPolicy(mode="sync"),
        "worai_help": ToolPolicy(mode="sync"),
        "worai_graph_audit": ToolPolicy(mode="sync"),
        "worai_graph_reset": ToolPolicy(mode="sync"),
        "worai_graph_validate": ToolPolicy(mode="sync"),
        "worai_graph_export": ToolPolicy(mode="sync"),
        "worai_entity_matrix": ToolPolicy(mode="sync"),
        "worai_structured_data_validate_page": ToolPolicy(mode="sync"),
        "worai_web_pages_classify_types": ToolPolicy(mode="sync"),
        "worai_delete_entities_from_csv": ToolPolicy(mode="sync"),
        "worai_canonicalize_duplicate_pages": ToolPolicy(mode="sync"),
        "worai_graph_property_delete": ToolPolicy(mode="sync"),
        "worai_cli_exec": ToolPolicy(mode="sync"),
        "worai_structured_data_inventory_start": ToolPolicy(mode="async_required"),
        "worai_graph_sync_run_start": ToolPolicy(mode="async_required"),
        "worai_job_status": ToolPolicy(mode="sync"),
        "worai_job_output": ToolPolicy(mode="sync"),
        "worai_job_events": ToolPolicy(mode="sync"),
        "worai_job_cancel": ToolPolicy(mode="sync"),
    }

    def __init__(self, defaults: SessionDefaults) -> None:
        self._defaults = defaults
        self._runner = CliRunner()
        self._executor = WoraiExecutor(defaults=defaults, runner=self._runner)
        self._jobs = WoraiJobManager()

    def serve(self) -> int:
        while True:
            message = McpTransport.read_message()
            if message is None:
                return 0
            response = self.handle_message(message)
            if response is not None:
                McpTransport.write_message(response)

    def handle_message(self, message: dict[str, Any]) -> dict[str, Any] | None:
        if "method" not in message:
            return None

        method = str(message["method"])
        request_id = message.get("id")
        params = message.get("params") or {}
        if not isinstance(params, dict):
            params = {}

        if request_id is None:
            if method == "notifications/initialized":
                return None
            return None
        try:
            result = self._dispatch(method, params)
            return {"jsonrpc": _JSONRPC, "id": request_id, "result": result}
        except Exception as exc:  # pragma: no cover - defensive branch
            return {
                "jsonrpc": _JSONRPC,
                "id": request_id,
                "error": {"code": -32000, "message": str(exc)},
            }

    def _dispatch(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        if method == "initialize":
            return {
                "protocolVersion": "2024-11-05",
                "capabilities": {"tools": {}},
                "serverInfo": {"name": "worai-mcp", "version": "1.1.0"},
            }
        if method == "ping":
            return {}
        if method == "tools/list":
            return {"tools": self._tools()}
        if method == "tools/call":
            name = str(params.get("name", "")).strip()
            arguments = params.get("arguments") or {}
            if not isinstance(arguments, dict):
                arguments = {}
            payload = self._call_tool(name, arguments)
            return {
                "structuredContent": payload,
                "content": [{"type": "text", "text": json.dumps(payload)}],
                "isError": False,
            }
        raise ValueError(f"Method not found: {method}")

    def _tools(self) -> list[dict[str, Any]]:
        base_schema = {
            "config_path": {"type": "string"},
            "profile": {"type": "string"},
            "max_output_chars": {"type": "integer", "minimum": 256},
        }
        return [
            {
                "name": "worai_version",
                "description": "Show installed worai CLI version.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {"type": "object", "properties": base_schema, "additionalProperties": False},
            },
            {
                "name": "worai_help",
                "description": "Show worai --help for root or specific subcommand path.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "command_path": {"type": "array", "items": {"type": "string"}},
                    },
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_graph_audit",
                "description": (
                    "Run `worai graph audit` on a local graph file (TTL, JSON-LD, NT, RDF/XML). "
                    "Returns a schema compliance, rich-snippets, and connectivity report. "
                    "Exit code is non-zero when the file has load errors."
                ),
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "file": {"type": "string"},
                        "depth": {"type": "integer", "minimum": 0},
                        "list_isolated_iris": {"type": "boolean"},
                        "show_url_violations": {"type": "boolean"},
                        "rich_snippets_granularity": {"type": "string", "enum": ["counts", "entities"]},
                        "workers": {"type": "integer", "minimum": 1},
                        "builtin_shapes": {"type": "array", "items": {"type": "string"}},
                        "exclude_builtin_shapes": {"type": "array", "items": {"type": "string"}},
                        "shapes": {"type": "array", "items": {"type": "string"}},
                        "issue_level": {"type": "string", "enum": ["warning", "error"]},
                        "format": {"type": "string", "enum": ["text", "json"]},
                    },
                    "required": ["file"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_graph_reset",
                "description": (
                    "Run `worai graph reset` — resets the WordLift account associated with the "
                    "current profile. This operation is DESTRUCTIVE and IRREVERSIBLE. "
                    "Optionally keeps country, language, and/or URL settings."
                ),
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "keep_country": {"type": "boolean"},
                        "keep_language": {"type": "boolean"},
                        "keep_url": {"type": "boolean"},
                    },
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_graph_validate",
                "description": "Run `worai graph validate` against files/URLs.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "inputs": {"type": "array", "items": {"type": "string"}},
                        "builtin_shapes": {"type": "array", "items": {"type": "string"}},
                        "exclude_builtin_shapes": {"type": "array", "items": {"type": "string"}},
                        "shapes": {"type": "array", "items": {"type": "string"}},
                        "level": {"type": "string", "enum": ["warning", "error"]},
                        "format": {"type": "string", "enum": ["text", "json"]},
                    },
                    "required": ["inputs"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_graph_export",
                "description": "Run `worai graph export` with optional output file and --validate.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {**base_schema, "output_file": {"type": "string"}, "validate": {"type": "boolean"}},
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_entity_matrix",
                "description": (
                    "Run `worai entity-matrix` on a local TTL graph file. "
                    "Returns a URL × entity-type pivot table: one row per URL (or cluster), "
                    "one column per entity-type short name, cell values are integer counts. "
                    "Supports optional type exclusion and automatic URL clustering."
                ),
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "file": {"type": "string"},
                        "exclude_types": {"type": "array", "items": {"type": "string"}},
                        "cluster": {"type": "boolean"},
                        "depth": {"type": "integer", "minimum": 1},
                        "format": {
                            "type": "string",
                            "enum": ["table", "json", "csv", "tsv"],
                        },
                        "output_file": {"type": "string"},
                    },
                    "required": ["file"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_structured_data_validate_page",
                "description": "Run `worai structured-data validate page <url>`.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "url": {"type": "string"},
                        "shapes": {"type": "array", "items": {"type": "string"}},
                        "format": {"type": "string", "enum": ["pretty", "raw"]},
                        "color": {"type": "boolean"},
                        "headed": {"type": "boolean"},
                        "timeout_ms": {"type": "integer"},
                    },
                    "required": ["url"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_web_pages_classify_types",
                "description": "Run `worai web-pages classify-types` for a source and output CSV.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "source": {"type": "string"},
                        "output_csv": {"type": "string"},
                        "ingest_source": {"type": "string"},
                        "ingest_loader": {"type": "string"},
                        "url_regex": {"type": "string"},
                        "agent_cli": {"type": "string", "enum": ["codex", "claude", "gemini"]},
                        "agent_timeout_sec": {"type": "number"},
                        "max_markdown_chars": {"type": "integer"},
                    },
                    "required": ["source"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_delete_entities_from_csv",
                "description": (
                    "Run `worai delete-entities-from-csv` to delete entities whose IRIs are listed "
                    "in the first column of a CSV file. DESTRUCTIVE and IRREVERSIBLE."
                ),
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "csv_file": {"type": "string"},
                        "batch_size": {"type": "integer", "minimum": 1},
                    },
                    "required": ["csv_file"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_canonicalize_duplicate_pages",
                "description": (
                    "Run `worai canonicalize-duplicate-pages` — reads a GSC CSV produced by "
                    "`worai google-search-console`, queries the WordLift graph to find the "
                    "canonical target for each duplicate URL, and writes results to an output CSV."
                ),
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "input_csv": {"type": "string"},
                        "output_csv": {"type": "string"},
                        "url_regex": {"type": "string"},
                        "entity_type": {"type": "string"},
                        "endpoint": {"type": "string"},
                        "batch_size": {"type": "integer", "minimum": 1},
                        "kpi_window": {"type": "string", "enum": ["7d", "28d", "3m"]},
                        "kpi_metric": {"type": "string", "enum": ["clicks", "impressions", "ctr"]},
                    },
                    "required": ["input_csv"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_graph_property_delete",
                "description": (
                    "Run `worai graph property delete` — deletes one predicate from all matching "
                    "entities in the WordLift graph. DESTRUCTIVE. Use dry_run=true first to preview "
                    "matching entities without applying changes."
                ),
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "predicate": {"type": "string"},
                        "endpoint": {"type": "string"},
                        "entities_endpoint": {"type": "string"},
                        "dry_run": {"type": "boolean"},
                        "workers": {"type": "integer", "minimum": 1},
                        "retries": {"type": "integer", "minimum": 0},
                        "rate_delay": {"type": "number", "minimum": 0},
                        "limit": {"type": "integer", "minimum": 0},
                        "timeout": {"type": "integer", "minimum": 1},
                    },
                    "required": ["predicate"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_structured_data_inventory_start",
                "description": "Start `worai structured-data inventory` as a background job.",
                "annotations": {"x-execution-mode": "async_required"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "source": {"type": "string"},
                        "sheet_name": {"type": "string"},
                        "url_regex": {"type": "string"},
                        "source_type": {"type": "string"},
                        "ingest_source": {"type": "string"},
                        "ingest_loader": {"type": "string"},
                        "ingest_passthrough_when_html": {"type": "boolean"},
                        "output": {"type": "string"},
                        "destination_sheet_id": {"type": "string"},
                        "destination_sheet_name": {"type": "string"},
                        "client_secrets": {"type": "string"},
                        "token": {"type": "string"},
                        "port": {"type": "integer"},
                        "timeout": {"type": "number"},
                        "concurrency": {"type": "string"},
                    },
                    "required": ["source"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_graph_sync_run_start",
                "description": "Start `worai graph sync run` as a background job.",
                "annotations": {"x-execution-mode": "async_required"},
                "inputSchema": {
                    "type": "object",
                    "properties": {**base_schema, "debug": {"type": "boolean"}},
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_job_status",
                "description": "Get status for an async worai job.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {"job_id": {"type": "string"}},
                    "required": ["job_id"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_job_output",
                "description": "Read incremental output for an async worai job.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        "job_id": {"type": "string"},
                        "offset": {"type": "integer", "minimum": 0},
                        "max_output_chars": {"type": "integer", "minimum": 256},
                    },
                    "required": ["job_id"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_job_events",
                "description": "Read structured progress events for an async worai job.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        "job_id": {"type": "string"},
                        "cursor": {"type": "integer", "minimum": 0},
                        "limit": {"type": "integer", "minimum": 1},
                    },
                    "required": ["job_id"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_job_cancel",
                "description": "Cancel an async worai job.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {"job_id": {"type": "string"}},
                    "required": ["job_id"],
                    "additionalProperties": False,
                },
            },
            {
                "name": "worai_cli_exec",
                "description": "Fallback executor for uncovered commands. Accepts argv array only.",
                "annotations": {"x-execution-mode": "sync"},
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        **base_schema,
                        "argv": {"type": "array", "items": {"type": "string"}},
                    },
                    "required": ["argv"],
                    "additionalProperties": False,
                },
            },
        ]

    def _call_tool(self, name: str, arguments: dict[str, Any]) -> dict[str, Any]:
        policy = self.TOOL_POLICIES.get(name)
        if policy is None:
            raise ValueError(f"Unknown tool: {name}")

        if name == "worai_version":
            return self._invoke_worai_sync(
                argv=["version"],
                config_path=self._optional_str(arguments.get("config_path")),
                profile=self._optional_str(arguments.get("profile")),
                max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
            )
        if name == "worai_help":
            return self._tool_worai_help(arguments)
        if name == "worai_graph_audit":
            return self._tool_worai_graph_audit(arguments)
        if name == "worai_graph_reset":
            return self._tool_worai_graph_reset(arguments)
        if name == "worai_graph_validate":
            return self._tool_worai_graph_validate(arguments)
        if name == "worai_graph_export":
            return self._tool_worai_graph_export(arguments)
        if name == "worai_entity_matrix":
            return self._tool_worai_entity_matrix(arguments)
        if name == "worai_structured_data_validate_page":
            return self._tool_worai_structured_data_validate_page(arguments)
        if name == "worai_web_pages_classify_types":
            return self._tool_worai_web_pages_classify_types(arguments)
        if name == "worai_delete_entities_from_csv":
            return self._tool_worai_delete_entities_from_csv(arguments)
        if name == "worai_canonicalize_duplicate_pages":
            return self._tool_worai_canonicalize_duplicate_pages(arguments)
        if name == "worai_graph_property_delete":
            return self._tool_worai_graph_property_delete(arguments)
        if name == "worai_cli_exec":
            return self._tool_worai_cli_exec(arguments)
        if name == "worai_structured_data_inventory_start":
            return self._tool_worai_structured_data_inventory_start(arguments)
        if name == "worai_graph_sync_run_start":
            return self._tool_worai_graph_sync_run_start(arguments)
        if name == "worai_job_status":
            job_id = self._required_str(arguments.get("job_id"), field_name="job_id")
            return self._jobs.status(job_id)
        if name == "worai_job_output":
            job_id = self._required_str(arguments.get("job_id"), field_name="job_id")
            offset = int(arguments.get("offset", 0) or 0)
            max_chars = self._max_output_chars(arguments.get("max_output_chars"))
            return self._jobs.output(job_id, offset=offset, max_chars=max_chars)
        if name == "worai_job_events":
            job_id = self._required_str(arguments.get("job_id"), field_name="job_id")
            cursor = int(arguments.get("cursor", 0) or 0)
            limit = int(arguments.get("limit", 100) or 100)
            return self._jobs.events(job_id, cursor=cursor, limit=limit)
        if name == "worai_job_cancel":
            job_id = self._required_str(arguments.get("job_id"), field_name="job_id")
            return self._jobs.cancel(job_id)
        raise ValueError(f"Unknown tool: {name}")

    def _invoke_worai_sync(
        self,
        *,
        argv: list[str],
        config_path: str | None,
        profile: str | None,
        max_output_chars: int,
    ) -> dict[str, Any]:
        started = time.time()
        result = self._executor.invoke_sync(
            argv=argv,
            config_path=config_path,
            profile=profile,
            max_output_chars=max_output_chars,
        )
        elapsed = time.time() - started
        if elapsed > _SYNC_TIMEOUT_SEC:
            raise ValueError("Sync execution exceeded threshold. Use async *_start tool.")
        return result

    def _tool_worai_help(self, arguments: dict[str, Any]) -> dict[str, Any]:
        raw_command_path = arguments.get("command_path") or []
        if raw_command_path and (
            not isinstance(raw_command_path, list)
            or not all(isinstance(item, str) and item.strip() for item in raw_command_path)
        ):
            raise ValueError("command_path must be an array of non-empty strings")
        argv = [*raw_command_path, "--help"]
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_delete_entities_from_csv(self, arguments: dict[str, Any]) -> dict[str, Any]:
        csv_file = self._required_str(arguments.get("csv_file"), field_name="csv_file")
        argv: list[str] = ["delete-entities-from-csv", csv_file]
        batch_size = arguments.get("batch_size")
        if batch_size is not None:
            argv.extend(["--batch-size", str(int(batch_size))])
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_canonicalize_duplicate_pages(self, arguments: dict[str, Any]) -> dict[str, Any]:
        input_csv = self._required_str(arguments.get("input_csv"), field_name="input_csv")
        argv: list[str] = ["canonicalize-duplicate-pages", "--input", input_csv]
        output_csv = self._optional_str(arguments.get("output_csv"))
        if output_csv:
            argv.extend(["--output", output_csv])
        url_regex = self._optional_str(arguments.get("url_regex"))
        if url_regex:
            argv.extend(["--url-regex", url_regex])
        entity_type = self._optional_str(arguments.get("entity_type"))
        if entity_type:
            argv.extend(["--entity-type", entity_type])
        endpoint = self._optional_str(arguments.get("endpoint"))
        if endpoint:
            argv.extend(["--endpoint", endpoint])
        batch_size = arguments.get("batch_size")
        if batch_size is not None:
            argv.extend(["--batch-size", str(int(batch_size))])
        kpi_window = self._optional_str(arguments.get("kpi_window"))
        if kpi_window:
            argv.extend(["--kpi-window", kpi_window])
        kpi_metric = self._optional_str(arguments.get("kpi_metric"))
        if kpi_metric:
            argv.extend(["--kpi-metric", kpi_metric])
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_graph_property_delete(self, arguments: dict[str, Any]) -> dict[str, Any]:
        predicate = self._required_str(arguments.get("predicate"), field_name="predicate")
        argv: list[str] = ["graph", "property", "delete", predicate, "--yes"]
        if bool(arguments.get("dry_run")):
            argv.append("--dry-run")
        endpoint = self._optional_str(arguments.get("endpoint"))
        if endpoint:
            argv.extend(["--endpoint", endpoint])
        entities_endpoint = self._optional_str(arguments.get("entities_endpoint"))
        if entities_endpoint:
            argv.extend(["--entities-endpoint", entities_endpoint])
        workers = arguments.get("workers")
        if workers is not None:
            argv.extend(["--workers", str(int(workers))])
        retries = arguments.get("retries")
        if retries is not None:
            argv.extend(["--retries", str(int(retries))])
        rate_delay = arguments.get("rate_delay")
        if rate_delay is not None:
            argv.extend(["--rate-delay", str(float(rate_delay))])
        limit = arguments.get("limit")
        if limit is not None:
            argv.extend(["--limit", str(int(limit))])
        timeout = arguments.get("timeout")
        if timeout is not None:
            argv.extend(["--timeout", str(int(timeout))])
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_cli_exec(self, arguments: dict[str, Any]) -> dict[str, Any]:
        raw_argv = arguments.get("argv")
        if not isinstance(raw_argv, list) or not raw_argv:
            raise ValueError("argv must be a non-empty array of strings")
        if not all(isinstance(item, str) and item.strip() for item in raw_argv):
            raise ValueError("argv must contain only non-empty strings")
        argv = [item.strip() for item in raw_argv]
        for blocked in _FORBIDDEN_ARGV_PREFIXES:
            if tuple(argv[: len(blocked)]) == blocked:
                raise ValueError(f"Command is blocked: {' '.join(blocked)}")
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_graph_audit(self, arguments: dict[str, Any]) -> dict[str, Any]:
        file = self._required_str(arguments.get("file"), field_name="file")
        argv: list[str] = ["graph", "audit", file]
        depth = arguments.get("depth")
        if depth is not None:
            argv.extend(["--depth", str(int(depth))])
        if bool(arguments.get("list_isolated_iris")):
            argv.append("--list-isolated-iris")
        if bool(arguments.get("show_url_violations")):
            argv.append("--show-url-violations")
        granularity = self._optional_str(arguments.get("rich_snippets_granularity"))
        if granularity:
            argv.extend(["--rich-snippets-granularity", granularity])
        workers = arguments.get("workers")
        if workers is not None:
            argv.extend(["--workers", str(int(workers))])
        for shape in self._string_list(arguments.get("builtin_shapes"), field_name="builtin_shapes"):
            argv.extend(["--builtin-shape", shape])
        for shape in self._string_list(arguments.get("exclude_builtin_shapes"), field_name="exclude_builtin_shapes"):
            argv.extend(["--exclude-builtin-shape", shape])
        for shape in self._string_list(arguments.get("shapes"), field_name="shapes"):
            argv.extend(["--shape", shape])
        issue_level = self._optional_str(arguments.get("issue_level"))
        if issue_level:
            argv.extend(["--issue-level", issue_level])
        output_format = self._optional_str(arguments.get("format"))
        if output_format:
            argv.extend(["--format", output_format])
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_graph_reset(self, arguments: dict[str, Any]) -> dict[str, Any]:
        argv: list[str] = ["graph", "reset", "--yes"]
        if arguments.get("keep_country"):
            argv.append("--keep-country")
        if arguments.get("keep_language"):
            argv.append("--keep-language")
        if arguments.get("keep_url"):
            argv.append("--keep-url")
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_graph_validate(self, arguments: dict[str, Any]) -> dict[str, Any]:
        inputs = self._string_list(arguments.get("inputs"), field_name="inputs", required=True)
        argv: list[str] = ["graph", "validate", *inputs]
        for shape in self._string_list(arguments.get("builtin_shapes"), field_name="builtin_shapes"):
            argv.extend(["--builtin-shape", shape])
        for shape in self._string_list(arguments.get("exclude_builtin_shapes"), field_name="exclude_builtin_shapes"):
            argv.extend(["--exclude-builtin-shape", shape])
        for shape in self._string_list(arguments.get("shapes"), field_name="shapes"):
            argv.extend(["--shape", shape])
        level = self._optional_str(arguments.get("level"))
        if level:
            argv.extend(["--level", level])
        output_format = self._optional_str(arguments.get("format"))
        if output_format:
            argv.extend(["--format", output_format])
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_graph_export(self, arguments: dict[str, Any]) -> dict[str, Any]:
        argv: list[str] = ["graph", "export"]
        output_file = self._optional_str(arguments.get("output_file"))
        if output_file:
            argv.append(output_file)
        if bool(arguments.get("validate")):
            argv.append("--validate")
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_entity_matrix(self, arguments: dict[str, Any]) -> dict[str, Any]:
        file = self._required_str(arguments.get("file"), field_name="file")
        argv: list[str] = ["entity-matrix", file]
        for t in self._string_list(arguments.get("exclude_types"), field_name="exclude_types"):
            argv.extend(["--exclude-type", t])
        cluster = arguments.get("cluster")
        if cluster is True:
            argv.append("--cluster")
        elif cluster is False:
            argv.append("--no-cluster")
        depth = arguments.get("depth")
        if depth is not None:
            argv.extend(["--depth", str(int(depth))])
        output_format = self._optional_str(arguments.get("format"))
        if output_format:
            argv.extend(["--format", output_format])
        output_file = self._optional_str(arguments.get("output_file"))
        if output_file:
            argv.extend(["--output", output_file])
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_structured_data_validate_page(self, arguments: dict[str, Any]) -> dict[str, Any]:
        url = self._required_str(arguments.get("url"), field_name="url")
        argv: list[str] = ["structured-data", "validate", "page", url]
        for shape in self._string_list(arguments.get("shapes"), field_name="shapes"):
            argv.extend(["--shape", shape])
        output_format = self._optional_str(arguments.get("format"))
        if output_format:
            argv.extend(["--format", output_format])
        color = arguments.get("color")
        if color is True:
            argv.append("--color")
        if color is False:
            argv.append("--no-color")
        if bool(arguments.get("headed")):
            argv.append("--headed")
        timeout_ms = arguments.get("timeout_ms")
        if timeout_ms is not None:
            argv.extend(["--timeout-ms", str(int(timeout_ms))])
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_web_pages_classify_types(self, arguments: dict[str, Any]) -> dict[str, Any]:
        source = self._required_str(arguments.get("source"), field_name="source")
        argv: list[str] = ["web-pages", "classify-types", source, "--yes"]
        output_csv = self._optional_str(arguments.get("output_csv"))
        if output_csv:
            argv.extend(["--output", output_csv])
        ingest_source = self._optional_str(arguments.get("ingest_source"))
        if ingest_source:
            argv.extend(["--ingest-source", ingest_source])
        ingest_loader = self._optional_str(arguments.get("ingest_loader"))
        if ingest_loader:
            argv.extend(["--ingest-loader", ingest_loader])
        url_regex = self._optional_str(arguments.get("url_regex"))
        if url_regex:
            argv.extend(["--url-regex", url_regex])
        agent_cli = self._optional_str(arguments.get("agent_cli"))
        if agent_cli:
            argv.extend(["--agent-cli", agent_cli])
        timeout = arguments.get("agent_timeout_sec")
        if timeout is not None:
            argv.extend(["--agent-timeout-sec", str(float(timeout))])
        max_markdown = arguments.get("max_markdown_chars")
        if max_markdown is not None:
            argv.extend(["--max-markdown-chars", str(int(max_markdown))])
        return self._invoke_worai_sync(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
            max_output_chars=self._max_output_chars(arguments.get("max_output_chars")),
        )

    def _tool_worai_structured_data_inventory_start(self, arguments: dict[str, Any]) -> dict[str, Any]:
        source = self._required_str(arguments.get("source"), field_name="source")
        argv: list[str] = ["structured-data", "inventory", source]
        for arg_name, option_name in (
            ("sheet_name", "--sheet-name"),
            ("url_regex", "--url-regex"),
            ("source_type", "--source-type"),
            ("ingest_source", "--ingest-source"),
            ("ingest_loader", "--ingest-loader"),
            ("output", "--output"),
            ("destination_sheet_id", "--destination-sheet-id"),
            ("destination_sheet_name", "--destination-sheet-name"),
            ("client_secrets", "--client-secrets"),
            ("token", "--token"),
            ("concurrency", "--concurrency"),
        ):
            value = self._optional_str(arguments.get(arg_name))
            if value:
                argv.extend([option_name, value])

        passthrough = arguments.get("ingest_passthrough_when_html")
        if passthrough is True:
            argv.append("--ingest-passthrough-when-html")
        if passthrough is False:
            argv.append("--no-ingest-passthrough-when-html")

        timeout = arguments.get("timeout")
        if timeout is not None:
            argv.extend(["--timeout", str(float(timeout))])
        port = arguments.get("port")
        if port is not None:
            argv.extend(["--port", str(int(port))])

        command, resolved_config, effective_profile = self._executor.build_subprocess_command(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
        )
        return self._jobs.start(
            command=command,
            effective_config_path=resolved_config,
            effective_profile=effective_profile,
        )

    def _tool_worai_graph_sync_run_start(self, arguments: dict[str, Any]) -> dict[str, Any]:
        argv: list[str] = ["graph", "sync", "run"]
        if bool(arguments.get("debug")):
            argv.append("--debug")
        command, resolved_config, effective_profile = self._executor.build_subprocess_command(
            argv=argv,
            config_path=self._optional_str(arguments.get("config_path")),
            profile=self._optional_str(arguments.get("profile")),
        )
        return self._jobs.start(
            command=command,
            effective_config_path=resolved_config,
            effective_profile=effective_profile,
        )

    @staticmethod
    def _optional_str(value: Any) -> str | None:
        if value is None:
            return None
        if not isinstance(value, str):
            raise ValueError("Expected a string value")
        stripped = value.strip()
        return stripped or None

    @staticmethod
    def _required_str(value: Any, *, field_name: str) -> str:
        optional = WoraiMcpServer._optional_str(value)
        if not optional:
            raise ValueError(f"{field_name} is required")
        return optional

    @staticmethod
    def _string_list(value: Any, *, field_name: str, required: bool = False) -> list[str]:
        if value is None:
            if required:
                raise ValueError(f"{field_name} is required")
            return []
        if not isinstance(value, list):
            raise ValueError(f"{field_name} must be an array of non-empty strings")
        if not all(isinstance(item, str) and item.strip() for item in value):
            raise ValueError(f"{field_name} must be an array of non-empty strings")
        return [item.strip() for item in value]

    @staticmethod
    def _max_output_chars(value: Any) -> int:
        if value is None:
            return _MAX_OUTPUT_CHARS_DEFAULT
        try:
            parsed = int(value)
        except (TypeError, ValueError) as exc:
            raise ValueError("max_output_chars must be an integer") from exc
        return max(256, min(parsed, _MAX_OUTPUT_CHARS_LIMIT))


def _truncate(text: str, max_chars: int) -> tuple[str, bool]:
    if len(text) <= max_chars:
        return text, False
    return text[:max_chars], True


def run_stdio_server(*, config_path: str | None, profile: str | None) -> int:
    defaults = SessionDefaults(config_path=config_path, profile=profile)
    return WoraiMcpServer(defaults=defaults).serve()


@dataclass
class RunningHttpServer:
    server: ThreadingHTTPServer
    thread: threading.Thread
    url: str

    def stop(self) -> None:
        self.server.shutdown()
        self.server.server_close()
        self.thread.join(timeout=2.0)


class _HttpMcpHandler(BaseHTTPRequestHandler):
    protocol_version = "HTTP/1.1"

    @property
    def _mcp(self) -> WoraiMcpServer:
        return self.server.mcp_server  # type: ignore[attr-defined]

    def _send_json(self, status: int, payload: dict[str, Any]) -> None:
        encoded = json.dumps(payload).encode("utf-8")
        self.send_response(status)
        self.send_header("Content-Type", "application/json")
        self.send_header("Content-Length", str(len(encoded)))
        self.end_headers()
        self.wfile.write(encoded)

    def do_GET(self) -> None:  # noqa: N802
        if self.path == "/health":
            self._send_json(200, {"ok": True})
            return
        self.send_response(404)
        self.send_header("Content-Length", "0")
        self.end_headers()

    def do_POST(self) -> None:  # noqa: N802
        if self.path != "/mcp":
            self.send_response(404)
            self.send_header("Content-Length", "0")
            self.end_headers()
            return
        try:
            content_length = int(self.headers.get("Content-Length", "0"))
        except ValueError:
            self._send_json(400, {"error": "invalid content length"})
            return
        body = self.rfile.read(content_length) if content_length > 0 else b""
        try:
            data = json.loads(body.decode("utf-8"))
        except json.JSONDecodeError:
            self._send_json(400, {"error": "invalid json"})
            return
        if not isinstance(data, dict):
            self._send_json(400, {"error": "json-rpc object required"})
            return
        response = self._mcp.handle_message(data)
        if response is None:
            self.send_response(202)
            self.send_header("Content-Length", "0")
            self.end_headers()
            return
        self._send_json(200, response)

    def log_message(self, format: str, *args: object) -> None:  # noqa: A003
        return


def start_http_server(
    *,
    host: str,
    port: int,
    config_path: str | None,
    profile: str | None,
) -> RunningHttpServer:
    defaults = SessionDefaults(config_path=config_path, profile=profile)
    mcp_server = WoraiMcpServer(defaults=defaults)
    http = ThreadingHTTPServer((host, int(port)), _HttpMcpHandler)
    http.mcp_server = mcp_server  # type: ignore[attr-defined]
    thread = threading.Thread(target=http.serve_forever, daemon=True)
    thread.start()
    actual_port = int(http.server_address[1])
    return RunningHttpServer(
        server=http,
        thread=thread,
        url=f"http://{host}:{actual_port}/mcp",
    )


def _parse_args(argv: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Run worai MCP server over stdio.")
    parser.add_argument("--config", dest="config_path", default=None)
    parser.add_argument("--profile", dest="profile", default=None)
    return parser.parse_args(argv)


def main(argv: list[str] | None = None) -> int:
    args = _parse_args(argv or sys.argv[1:])
    resolved_config = str(resolve_config_path(args.config_path, env=os.environ)) if args.config_path else None
    return run_stdio_server(config_path=resolved_config, profile=args.profile)


if __name__ == "__main__":
    raise SystemExit(main())



"""Graph sync workflow integration."""

from __future__ import annotations

import asyncio
import ast
import json
import logging
import os
import re
import tomllib
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from worai.core.ingest import DEFAULT_INGEST_LOADER, resolve_ingest_settings

try:
    from wordlift_sdk.configuration import ConfigurationProvider
    from wordlift_sdk.kg_build import (
        ProfileImportProtocol,
        load_profile_config,
    )
    from wordlift_sdk.kg_build.cloud_flow import CloudWorkflowConfig, run_cloud_workflow
    from wordlift_sdk.kg_build.container import KgBuildApplicationContainer
except ModuleNotFoundError:  # pragma: no cover - environment dependent
    ConfigurationProvider = None
    ProfileImportProtocol = None
    load_profile_config = None
    CloudWorkflowConfig = None
    run_cloud_workflow = None
    KgBuildApplicationContainer = None


LOGGER = logging.getLogger("worai.graph")
_TIMEOUT_RE = re.compile(r"^(?P<value>\d+(?:\.\d+)?)(?P<unit>ms|s)?$")


@dataclass
class GraphSyncOptions:
    profile: str
    config_path: Path = Path("worai.toml")
    debug: bool = False


def load_service_account_json(raw_value: str | None) -> str:
    if not raw_value:
        raise ValueError("Missing profile setting: sheets_service_account")

    candidate_path = Path(raw_value)
    if candidate_path.exists() and candidate_path.is_file():
        return candidate_path.read_text(encoding="utf-8")

    try:
        parsed = json.loads(raw_value)
    except json.JSONDecodeError as exc:
        raise ValueError("sheets_service_account must be JSON content or a file path") from exc

    if not isinstance(parsed, dict):
        raise ValueError("sheets_service_account JSON must be an object")
    return raw_value


def build_extra_settings(profile_settings: dict[str, Any]) -> dict[str, Any]:
    consumed = {
        "urls",
        "sitemap_url",
        "sitemap_url_pattern",
        "sheets_url",
        "sheets_name",
        "sheets_service_account",
        "overwrite",
        "concurrency",
        "web_page_import_mode",
        "web_page_import_timeout",
        "ingest_timeout_ms",
        "ingest_playwright_wait_until",
        "playwright_wait_until",
        "ingest_source",
        "ingest_loader",
        "ingest_passthrough_when_html",
        "google_search_console",
        "shacl_validate_sync",
        "shacl_validate_mode",
        "shacl_shape_specs",
        "shacl_builtin_shapes",
        "shacl_exclude_builtin_shapes",
        "shacl_extra_shapes",
        "canonical_id_strategy",
    }
    return {key.upper(): value for key, value in profile_settings.items() if key not in consumed}


def _append_py_setting(lines: list[str], key: str, value: Any) -> None:
    lines.append(f"{key} = {repr(value)}")


def _using_sheets_source(settings: dict[str, Any]) -> bool:
    return bool(settings.get("sheets_url") or settings.get("sheets_name"))


def _parse_bool(value: Any, *, name: str) -> bool:
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        lowered = value.strip().lower()
        if lowered in {"1", "true", "yes", "on"}:
            return True
        if lowered in {"0", "false", "no", "off"}:
            return False
    raise ValueError(f"{name} must be a boolean")


def _resolve_web_page_import_timeout_ms(value: Any) -> int:
    if value is None:
        raise ValueError("web_page_import_timeout must not be empty")

    parsed: float
    explicit_ms = False
    explicit_s = False
    if isinstance(value, (int, float)):
        parsed = float(value)
    elif isinstance(value, str):
        token = value.strip().lower()
        if not token:
            raise ValueError("web_page_import_timeout must not be empty")
        match = _TIMEOUT_RE.match(token)
        if not match:
            raise ValueError(
                "web_page_import_timeout must be a number of seconds (e.g. 60 or 60s) "
                "or milliseconds (e.g. 60000ms)"
            )
        parsed = float(match.group("value"))
        explicit_ms = match.group("unit") == "ms"
        explicit_s = match.group("unit") == "s"
    else:
        raise ValueError(
            "web_page_import_timeout must be numeric seconds or a string with optional s/ms suffix"
        )

    if parsed <= 0:
        raise ValueError("web_page_import_timeout must be greater than 0")

    if explicit_ms:
        return int(parsed)
    if explicit_s:
        return int(parsed * 1000)

    return int(parsed * 1000)


def _resolve_playwright_wait_until(value: Any) -> str:
    if not isinstance(value, str):
        raise ValueError("playwright_wait_until must be a string")
    normalized = value.strip().lower()
    if normalized not in {"domcontentloaded", "load", "networkidle"}:
        raise ValueError("playwright_wait_until must be one of: domcontentloaded, load, networkidle")
    return normalized


def _resolve_canonical_id_strategy(value: Any) -> str:
    if not isinstance(value, str):
        raise ValueError("canonical_id_strategy must be a string")
    normalized = value.strip().lower()
    if normalized not in {"legacy", "dependency_graph"}:
        raise ValueError("canonical_id_strategy must be one of: legacy, dependency_graph")
    return normalized


def _resolve_ingest_timeout_ms(settings: dict[str, Any]) -> int | None:
    modern_value = settings.get("ingest_timeout_ms")
    if modern_value is not None:
        if isinstance(modern_value, bool):
            raise ValueError("ingest_timeout_ms must be an integer number of milliseconds")
        try:
            parsed = int(modern_value)
        except (TypeError, ValueError) as exc:
            raise ValueError(f"ingest_timeout_ms must be an integer number of milliseconds: {modern_value}") from exc
        if parsed <= 0:
            raise ValueError("ingest_timeout_ms must be greater than 0")
        return parsed
    legacy_value = settings.get("web_page_import_timeout")
    if legacy_value is not None:
        return _resolve_web_page_import_timeout_ms(legacy_value)
    return None


def _normalize_string_list(value: Any, *, name: str) -> list[str]:
    if value is None:
        return []
    if isinstance(value, str):
        item = value.strip()
        return [item] if item else []
    if isinstance(value, (list, tuple, set)):
        normalized: list[str] = []
        for item in value:
            if not isinstance(item, str):
                raise ValueError(f"{name} entries must be strings")
            stripped = item.strip()
            if stripped:
                normalized.append(stripped)
        return normalized
    raise ValueError(f"{name} must be a string or list of strings")


def _normalize_shacl_validate_mode(value: Any) -> str:
    if not isinstance(value, str):
        raise ValueError("shacl_validate_mode must be a string")
    mode = value.strip().lower()
    if mode in {"warn", "fail", "off"}:
        return mode
    raise ValueError("shacl_validate_mode must be one of: warn, fail, off")


def _extract_shacl_validation_settings(settings: dict[str, Any]) -> dict[str, Any]:
    resolved: dict[str, Any] = {}

    if "shacl_validate_sync" in settings and settings.get("shacl_validate_sync") is not None:
        raise ValueError("shacl_validate_sync is no longer supported. Use shacl_validate_mode.")
    if "shacl_shape_specs" in settings and settings.get("shacl_shape_specs") is not None:
        raise ValueError(
            "shacl_shape_specs is no longer supported. "
            "Use shacl_builtin_shapes/shacl_exclude_builtin_shapes/shacl_extra_shapes."
        )

    if "shacl_validate_mode" in settings and settings.get("shacl_validate_mode") is not None:
        resolved["SHACL_VALIDATE_MODE"] = _normalize_shacl_validate_mode(settings.get("shacl_validate_mode"))

    if "shacl_builtin_shapes" in settings and settings.get("shacl_builtin_shapes") is not None:
        resolved["SHACL_BUILTIN_SHAPES"] = _normalize_string_list(
            settings.get("shacl_builtin_shapes"),
            name="shacl_builtin_shapes",
        )
    if "shacl_exclude_builtin_shapes" in settings and settings.get("shacl_exclude_builtin_shapes") is not None:
        resolved["SHACL_EXCLUDE_BUILTIN_SHAPES"] = _normalize_string_list(
            settings.get("shacl_exclude_builtin_shapes"),
            name="shacl_exclude_builtin_shapes",
        )
    if "shacl_extra_shapes" in settings and settings.get("shacl_extra_shapes") is not None:
        resolved["SHACL_EXTRA_SHAPES"] = _normalize_string_list(
            settings.get("shacl_extra_shapes"),
            name="shacl_extra_shapes",
        )

    return resolved


def resolve_google_search_console(config_path: Path, profile_name: str) -> bool:
    try:
        data = tomllib.loads(config_path.read_text(encoding="utf-8"))
    except FileNotFoundError:
        return False
    except tomllib.TOMLDecodeError as exc:
        raise ValueError(f"Invalid TOML config: {config_path}: {exc}") from exc

    global_value = data.get("google_search_console", False)
    profile_table = data.get("profile") or data.get("profiles") or {}
    profile_value = profile_table.get(profile_name, {}).get("google_search_console")
    selected = profile_value if profile_value is not None else global_value
    return _parse_bool(selected, name="google_search_console")


def resolve_postprocessor_runtime(config_path: Path, profile_name: str) -> str | None:
    try:
        data = tomllib.loads(config_path.read_text(encoding="utf-8"))
    except FileNotFoundError:
        return None
    except tomllib.TOMLDecodeError as exc:
        raise ValueError(f"Invalid TOML config: {config_path}: {exc}") from exc

    profile_table = data.get("profile") or data.get("profiles") or {}
    profile_value = profile_table.get(profile_name, {}).get("postprocessor_runtime")
    global_value = data.get("postprocessor_runtime")
    selected = profile_value if profile_value is not None else global_value
    if selected is None:
        return None
    if not isinstance(selected, str):
        raise ValueError("postprocessor_runtime must be a string")

    runtime = selected.strip().lower()
    if runtime not in {"oneshot", "persistent"}:
        raise ValueError("postprocessor_runtime must be one of: oneshot, persistent")
    return runtime


def _build_dynamic_settings(
    profile_api_key: str,
    settings: dict[str, Any],
    *,
    google_search_console: bool,
) -> tuple[list[str], str | None]:
    lines = [f"WORDLIFT_KEY = {repr(profile_api_key)}"]
    sheets_service_account_json: str | None = None

    urls = settings.get("urls")
    sitemap_url = settings.get("sitemap_url")
    sitemap_url_pattern = settings.get("sitemap_url_pattern")
    sheets_url = settings.get("sheets_url")
    sheets_name = settings.get("sheets_name")
    using_sheets = _using_sheets_source(settings)
    source_modes = int(bool(urls)) + int(bool(sitemap_url)) + int(using_sheets)
    if source_modes != 1:
        raise ValueError(
            "Exactly one source is required: urls, sitemap_url, or sheets_url+sheets_name+sheets_service_account"
        )
    ingest = resolve_ingest_settings(
        context="graph_sync",
        new_source=settings.get("ingest_source"),
        new_loader=settings.get("ingest_loader"),
        new_passthrough_when_html=settings.get("ingest_passthrough_when_html"),
        legacy_loader=settings.get("web_page_import_mode"),
        default_loader=DEFAULT_INGEST_LOADER,
    )

    if urls:
        _append_py_setting(lines, "URLS", list(urls))
    elif sitemap_url:
        _append_py_setting(lines, "SITEMAP_URL", sitemap_url)
        if sitemap_url_pattern:
            _append_py_setting(lines, "SITEMAP_URL_PATTERN", sitemap_url_pattern)
    elif using_sheets:
        if not sheets_url or not sheets_name:
            raise ValueError("Both sheets_url and sheets_name are required when using Google Sheets source")
        sheets_service_account_json = load_service_account_json(settings.get("sheets_service_account"))
        _append_py_setting(lines, "SHEETS_URL", sheets_url)
        _append_py_setting(lines, "SHEETS_NAME", sheets_name)
    else:
        raise ValueError(
            "Exactly one source is required: urls, sitemap_url, or sheets_url+sheets_name+sheets_service_account"
        )

    source_value = ingest.source
    if source_value == "auto":
        if urls:
            source_value = "urls"
        elif sitemap_url:
            source_value = "sitemap"
        elif using_sheets:
            source_value = "sheets"
    _append_py_setting(lines, "INGEST_SOURCE", source_value)
    _append_py_setting(lines, "INGEST_LOADER", ingest.loader)
    _append_py_setting(lines, "INGEST_PASSTHROUGH_WHEN_HTML", ingest.passthrough_when_html)
    resolved_ingest_timeout_ms = _resolve_ingest_timeout_ms(settings)
    if resolved_ingest_timeout_ms is not None:
        _append_py_setting(lines, "INGEST_TIMEOUT_MS", resolved_ingest_timeout_ms)
    raw_wait_until = settings.get("ingest_playwright_wait_until", settings.get("playwright_wait_until"))
    if raw_wait_until is not None:
        _append_py_setting(
            lines,
            "PLAYWRIGHT_WAIT_UNTIL",
            _resolve_playwright_wait_until(raw_wait_until),
        )
    _append_py_setting(lines, "CONCURRENCY", int(settings.get("concurrency", 4)))
    _append_py_setting(lines, "OVERWRITE", bool(settings.get("overwrite", False)))
    _append_py_setting(lines, "GOOGLE_SEARCH_CONSOLE", google_search_console)
    for key, value in _extract_shacl_validation_settings(settings).items():
        _append_py_setting(lines, key, value)

    raw_canonical_id_strategy = settings.get("canonical_id_strategy")
    if raw_canonical_id_strategy is not None:
        _append_py_setting(
            lines,
            "CANONICAL_ID_STRATEGY",
            _resolve_canonical_id_strategy(raw_canonical_id_strategy),
        )

    for key, value in build_extra_settings(settings).items():
        _append_py_setting(lines, key, value)

    return lines, sheets_service_account_json


def _parse_settings_lines(lines: list[str]) -> dict[str, Any]:
    settings: dict[str, Any] = {}
    for line in lines:
        if "=" not in line:
            continue
        key, raw_value = line.split("=", 1)
        settings[key.strip()] = ast.literal_eval(raw_value.strip())
    return settings


def _build_cloud_workflow_config(
    *,
    lines: list[str],
    sheets_service_account_json: str | None,
    debug: bool,
    debug_profile_name: str,
) -> Any | None:
    if CloudWorkflowConfig is None:
        raise ValueError("wordlift-sdk 6.0.0+ cloud workflow config is unavailable in this environment")

    settings = _parse_settings_lines(lines)
    kwargs: dict[str, Any] = {
        "wordlift_key": settings.get("WORDLIFT_KEY"),
        "sheets_service_account_json": sheets_service_account_json,
        "overwrite": settings.get("OVERWRITE", False),
        "concurrency": settings.get("CONCURRENCY", 4),
        "ingest_loader": settings.get("INGEST_LOADER", "web_scrape_api"),
        "debug": debug,
        "debug_profile_name": debug_profile_name,
    }
    if "INGEST_TIMEOUT_MS" in settings:
        kwargs["ingest_timeout_ms"] = settings["INGEST_TIMEOUT_MS"]
    if "PLAYWRIGHT_WAIT_UNTIL" in settings:
        kwargs["playwright_wait_until"] = settings["PLAYWRIGHT_WAIT_UNTIL"]

    using_urls = "URLS" in settings
    using_sitemap = "SITEMAP_URL" in settings
    using_sheets = "SHEETS_URL" in settings or "SHEETS_NAME" in settings
    source_modes = int(using_urls) + int(using_sitemap) + int(using_sheets)
    if source_modes != 1:
        raise ValueError("Exactly one source mode is required: urls, sitemap_url, or sheets_url+sheets_name")

    if using_urls:
        kwargs["urls"] = settings["URLS"]
    elif using_sitemap:
        kwargs["sitemap_url"] = settings["SITEMAP_URL"]
        if "SITEMAP_URL_PATTERN" in settings:
            kwargs["sitemap_url_pattern"] = settings["SITEMAP_URL_PATTERN"]
    else:
        if "SHEETS_URL" not in settings or "SHEETS_NAME" not in settings:
            raise ValueError("Both sheets_url and sheets_name are required when using Google Sheets source")
        kwargs["sheets_url"] = settings["SHEETS_URL"]
        kwargs["sheets_name"] = settings["SHEETS_NAME"]

    consumed = {
        "WORDLIFT_KEY",
        "URLS",
        "SITEMAP_URL",
        "SITEMAP_URL_PATTERN",
        "SHEETS_URL",
        "SHEETS_NAME",
        "OVERWRITE",
        "CONCURRENCY",
        "INGEST_LOADER",
        "INGEST_TIMEOUT_MS",
        "PLAYWRIGHT_WAIT_UNTIL",
    }
    extra_settings = {key: value for key, value in settings.items() if key not in consumed}
    kwargs["extra_settings"] = extra_settings or None

    try:
        return CloudWorkflowConfig(**kwargs)
    except TypeError as exc:
        raise ValueError(
            "Installed wordlift-sdk does not match the expected 6.x cloud workflow contract."
        ) from exc


def _format_validation_summary(validation: Any) -> str:
    if validation is None:
        return "validation=disabled"
    if not isinstance(validation, dict):
        return f"validation={validation}"

    total = validation.get("total")
    passed = validation.get("pass")
    failed = validation.get("fail")
    warnings = (validation.get("warnings") or {}).get("count")
    errors = (validation.get("errors") or {}).get("count")
    return (
        f"validation(total={total}, pass={passed}, fail={failed}, "
        f"warnings={warnings}, errors={errors})"
    )


def _on_cloud_info(message: str) -> None:
    LOGGER.info(message)


def _on_cloud_progress(payload: dict[str, Any]) -> None:
    if not isinstance(payload, dict):
        LOGGER.info("Graph sync progress: %s", payload)
        return

    graph = payload.get("graph") or {}
    kind = payload.get("kind", "graph")
    location = payload.get("url") or payload.get("profile") or "n/a"
    LOGGER.info(
        "Progress[%s] %s entities=%s types=%s properties=%s %s",
        kind,
        location,
        graph.get("entities"),
        graph.get("type_assertions"),
        graph.get("property_assertions"),
        _format_validation_summary(payload.get("validation")),
    )
    LOGGER.debug("Progress payload: %s", json.dumps(payload, sort_keys=True, ensure_ascii=True))


def _on_cloud_kpi(payload: dict[str, Any]) -> None:
    if not isinstance(payload, dict):
        LOGGER.info("Graph sync KPI: %s", payload)
        return

    totals = payload.get("totals") or {}
    validation = payload.get("validation")
    LOGGER.info(
        "KPI profile=%s run_id=%s total_entities=%s type_assertions_total=%s property_assertions_total=%s",
        payload.get("profile"),
        payload.get("run_id"),
        totals.get("total_entities"),
        totals.get("type_assertions_total"),
        totals.get("property_assertions_total"),
    )
    LOGGER.info("KPI %s", _format_validation_summary(validation))
    if isinstance(validation, dict):
        total = validation.get("total")
        passed = validation.get("pass")
        failed = validation.get("fail")
        if isinstance(total, int) and isinstance(passed, int) and isinstance(failed, int):
            if total != passed + failed:
                LOGGER.warning(
                    "KPI validation invariant mismatch: total=%s pass=%s fail=%s",
                    total,
                    passed,
                    failed,
                )
    LOGGER.debug("KPI payload: %s", json.dumps(payload, sort_keys=True, ensure_ascii=True))


async def _run_cloud_workflow_with_callbacks(
    *,
    lines: list[str],
    sheets_service_account_json: str | None,
    debug: bool,
    debug_profile_name: str,
    protocol_factory,
) -> None:
    if (
        ConfigurationProvider is None
        or KgBuildApplicationContainer is None
        or run_cloud_workflow is None
    ):
        raise ValueError("wordlift-sdk 6.0.0+ cloud workflow support is unavailable in this environment")

    config = _build_cloud_workflow_config(
        lines=lines,
        sheets_service_account_json=sheets_service_account_json,
        debug=debug,
        debug_profile_name=debug_profile_name,
    )
    kwargs: dict[str, Any] = {
        "config": config,
        "configuration_provider_create": ConfigurationProvider.create,
        "container_factory": KgBuildApplicationContainer,
        "protocol_factory": protocol_factory,
        "on_info": _on_cloud_info,
        "on_progress": _on_cloud_progress,
        "on_kpi": _on_cloud_kpi,
    }
    await run_cloud_workflow(**kwargs)


async def run_graph_sync_async(options: GraphSyncOptions) -> None:
    if (
        ConfigurationProvider is None
        or ProfileImportProtocol is None
        or load_profile_config is None
        or KgBuildApplicationContainer is None
    ):
        raise ValueError("wordlift-sdk kg_build support is unavailable in this Python environment")

    config = load_profile_config(options.config_path)
    profile = config.get(options.profile)
    if profile is None:
        raise ValueError(f"Profile '{options.profile}' not found in {options.config_path}")
    if not profile.api_key:
        raise ValueError(f"Profile '{profile.name}' is missing api_key")

    settings = dict(profile.settings)
    google_search_console = resolve_google_search_console(options.config_path, options.profile)
    postprocessor_runtime = resolve_postprocessor_runtime(options.config_path, options.profile)
    lines, sheets_service_account_json = _build_dynamic_settings(
        profile.api_key,
        settings,
        google_search_console=google_search_console,
    )

    async def protocol_factory(context, debug_dir=None, workflow_config=None, on_progress=None):
        return ProfileImportProtocol(
            context=context,
            profile=profile,
            root_dir=Path.cwd(),
            debug_dir=debug_dir,
            on_progress=on_progress,
        )

    previous_postprocessor_runtime = os.environ.get("POSTPROCESSOR_RUNTIME")
    try:
        if postprocessor_runtime is not None:
            os.environ["POSTPROCESSOR_RUNTIME"] = postprocessor_runtime

        await _run_cloud_workflow_with_callbacks(
            lines=lines,
            sheets_service_account_json=sheets_service_account_json,
            debug=options.debug,
            debug_profile_name=profile.name,
            protocol_factory=protocol_factory,
        )
    finally:
        if postprocessor_runtime is not None:
            if previous_postprocessor_runtime is None:
                os.environ.pop("POSTPROCESSOR_RUNTIME", None)
            else:
                os.environ["POSTPROCESSOR_RUNTIME"] = previous_postprocessor_runtime


def run_graph_sync(options: GraphSyncOptions) -> None:
    asyncio.run(run_graph_sync_async(options))



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import pytest

from worai.core import google_oauth
from worai.errors import UsageError


class _Creds:
    def __init__(
        self,
        *,
        valid: bool = True,
        expired: bool = False,
        refresh_token: str | None = None,
        scopes: list[str] | None = None,
    ) -> None:
        self.valid = valid
        self.expired = expired
        self.refresh_token = refresh_token
        self.scopes = scopes or []
        self.refreshed = False

    def refresh(self, _request) -> None:
        self.refreshed = True
        self.valid = True

    def to_json(self) -> str:
        return '{"token":"x"}'


def test_load_google_credentials_uses_existing_valid_token(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    token = tmp_path / "token.json"
    token.write_text("{}", encoding="utf-8")
    creds = _Creds(valid=True, expired=False, scopes=["scope:a"])

    monkeypatch.setattr(google_oauth.os.path, "exists", lambda _p: True)
    monkeypatch.setattr(
        google_oauth.Credentials,
        "from_authorized_user_file",
        lambda _path, _scopes: creds,
    )
    monkeypatch.setattr(
        google_oauth.InstalledAppFlow,
        "from_client_secrets_file",
        lambda *_a, **_k: (_ for _ in ()).throw(AssertionError("flow should not run")),
    )

    result = google_oauth.load_google_credentials("client.json", str(token), 8080, ["scope:a"])
    assert result is creds
    assert token.read_text(encoding="utf-8") == '{"token":"x"}'


def test_load_google_credentials_refreshes_expired_token(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    token = tmp_path / "token.json"
    token.write_text("{}", encoding="utf-8")
    creds = _Creds(valid=True, expired=True, refresh_token="r", scopes=["scope:a"])

    monkeypatch.setattr(google_oauth.os.path, "exists", lambda _p: True)
    monkeypatch.setattr(google_oauth.Credentials, "from_authorized_user_file", lambda *_a, **_k: creds)

    result = google_oauth.load_google_credentials("client.json", str(token), 8080, ["scope:a"])
    assert result is creds
    assert creds.refreshed is True


def test_load_google_credentials_runs_oauth_flow_when_needed(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    token = tmp_path / "token.json"
    token.write_text("{}", encoding="utf-8")

    existing = _Creds(valid=True, expired=False, scopes=["scope:old"])
    fresh = _Creds(valid=True, expired=False, scopes=["scope:new"])
    calls: list[dict] = []

    monkeypatch.setattr(google_oauth.os.path, "exists", lambda _p: True)
    monkeypatch.setattr(google_oauth.Credentials, "from_authorized_user_file", lambda *_a, **_k: existing)

    def _flow_factory(_path, _scopes):
        def _run_local_server(**kwargs):
            calls.append(kwargs)
            return fresh

        return SimpleNamespace(run_local_server=_run_local_server)

    monkeypatch.setattr(google_oauth.InstalledAppFlow, "from_client_secrets_file", _flow_factory)

    result = google_oauth.load_google_credentials("client.json", str(token), 9090, ["scope:new"])
    assert result is fresh
    assert calls
    assert calls[0]["access_type"] == "offline"
    assert calls[0]["prompt"] == "consent"
    assert calls[0]["include_granted_scopes"] == "true"


def test_load_google_credentials_raises_without_client_secrets(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    token = tmp_path / "token.json"

    monkeypatch.setattr(google_oauth.os.path, "exists", lambda _p: False)

    with pytest.raises(UsageError, match="Google OAuth client secrets are required"):
        google_oauth.load_google_credentials(None, str(token), 8080, ["scope:a"])


def test_load_google_credentials_handles_invalid_stored_token_and_restores_env(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    token = tmp_path / "token.json"
    token.write_text("{}", encoding="utf-8")
    fresh = _Creds(valid=True, scopes=["scope:a"])

    monkeypatch.setenv("OAUTHLIB_RELAX_TOKEN_SCOPE", "0")
    monkeypatch.setattr(google_oauth.os.path, "exists", lambda _p: True)

    def _boom(*_a, **_k):
        raise ValueError("bad token")

    monkeypatch.setattr(google_oauth.Credentials, "from_authorized_user_file", _boom)
    monkeypatch.setattr(
        google_oauth.InstalledAppFlow,
        "from_client_secrets_file",
        lambda *_a, **_k: SimpleNamespace(run_local_server=lambda **_k2: fresh),
    )

    result = google_oauth.load_google_credentials("client.json", str(token), 8080, ["scope:a"])
    assert result is fresh
    assert google_oauth.os.environ["OAUTHLIB_RELAX_TOKEN_SCOPE"] == "0"



"""CLI wrapper for SHACL validation."""

from __future__ import annotations

from pathlib import Path
from typing import Iterable

import typer

from wordlift_sdk.validation import validate_file
from worai.errors import UsageError
from typer.core import TyperGroup
import click
from rdflib import BNode, Graph, Namespace, URIRef, RDF
from rdflib.collection import Collection

_SH = Namespace("http://www.w3.org/ns/shacl#")
_SCHEMA = "http://schema.org/"


def _shorten_term(term: URIRef) -> str:
    value = str(term)
    if value.startswith(_SCHEMA):
        return value[len(_SCHEMA) :]
    if "#" in value:
        return value.rsplit("#", 1)[-1]
    if "/" in value:
        return value.rsplit("/", 1)[-1]
    return value


def _path_to_str(graph: Graph, path: URIRef | BNode | None) -> str:
    if path is None:
        return ""
    if isinstance(path, BNode):
        items = [item for item in Collection(graph, path)]
        if items:
            return ".".join(_shorten_term(item) if isinstance(item, URIRef) else str(item) for item in items)
        return str(path)
    if isinstance(path, URIRef):
        return _shorten_term(path)
    return str(path)


def _describe_path(graph: Graph, path: URIRef | BNode | None) -> str:
    if path is None:
        return ""
    if isinstance(path, BNode):
        inv = graph.value(path, _SH.inversePath)
        if inv:
            return f"inverse { _shorten_term(inv) }"
        items = [item for item in Collection(graph, path)]
        if items:
            return ".".join(_shorten_term(item) if isinstance(item, URIRef) else str(item) for item in items)
    if isinstance(path, URIRef):
        return _shorten_term(path)
    return str(path)


def _or_shape_hint(graph: Graph, shape: URIRef | BNode | None) -> str | None:
    if shape is None:
        return None
    or_list = graph.value(shape, _SH["or"])
    if not or_list:
        return None
    alts = []
    for alt in Collection(graph, or_list):
        prop = graph.value(alt, _SH.property)
        if prop is None:
            continue
        path = graph.value(prop, _SH.path)
        if path is None:
            continue
        desc = _describe_path(graph, path)
        if desc:
            alts.append(desc)
    if not alts:
        return None
    if len(alts) == 1:
        return f"Provide: {alts[0]}."
    return f"Provide one of: {', '.join(alts)}."


def _format_pretty_lines(
    report_graph: Graph,
    data_graph: Graph | None,
    shape_source_map: dict[URIRef | BNode, str] | None,
) -> tuple[str, list[tuple[str, str]]]:
    results = []
    for result in report_graph.subjects(_SH.resultSeverity, None):
        severity = report_graph.value(result, _SH.resultSeverity)
        focus = report_graph.value(result, _SH.focusNode)
        path = report_graph.value(result, _SH.resultPath)
        message = report_graph.value(result, _SH.resultMessage)
        source_shape = report_graph.value(result, _SH.sourceShape)
        value_node = report_graph.value(result, _SH.value)
        results.append(
            {
                "severity": severity,
                "focus": focus,
                "path": path,
                "message": str(message) if message else "",
                "source_shape": source_shape,
                "value_node": value_node,
            }
        )

    if not results:
        return "Status: OK (0 errors, 0 warnings, 0 infos)", []

    def severity_label(sev: URIRef | None) -> str:
        if sev == _SH.Violation:
            return "ERROR"
        if sev == _SH.Warning:
            return "WARN"
        if sev == _SH.Info:
            return "INFO"
        return "WARN"

    errors = sum(1 for r in results if r["severity"] == _SH.Violation)
    warnings = sum(1 for r in results if r["severity"] == _SH.Warning)
    infos = sum(1 for r in results if r["severity"] == _SH.Info)

    status = "ERROR" if errors else "WARNING" if warnings else "INFO" if infos else "OK"
    lines: list[tuple[str, str]] = []
    header = f"Status: {status} ({errors} errors, {warnings} warnings, {infos} infos)"

    by_focus: dict[str, list[dict]] = {}
    for r in results:
        focus = str(r["focus"]) if r["focus"] is not None else "(unknown)"
        by_focus.setdefault(focus, []).append(r)

    for focus, items in by_focus.items():
        focus_types: list[str] = []
        if data_graph is not None and focus != "(unknown)":
            for t in data_graph.objects(URIRef(focus), RDF.type):
                focus_types.append(_shorten_term(t))
        if focus_types:
            lines.append(("text", f"Entity: {focus} (types: {', '.join(sorted(set(focus_types)))})"))
        else:
            lines.append(("text", f"Entity: {focus}"))
        for item in items:
            label = severity_label(item["severity"])
            path = _path_to_str(report_graph, item["path"])
            msg = item["message"]
            value_node = item.get("value_node")
            value_str = f" Value: {value_node}" if value_node is not None else ""
            focus = item.get("focus")
            focus_str = f" Focus: {focus}" if focus is not None else ""
            shape = item.get("source_shape")
            shape_str = ""
            if shape is not None:
                mapped = shape_source_map.get(shape) if shape_source_map else None
                if mapped:
                    shape_str = f" [shape: {mapped}]"
                elif isinstance(shape, URIRef):
                    shape_str = f" [shape: {_shorten_term(shape)}]"
                else:
                    shape_str = " [shape: (anonymous)]"
            google_required_suffix = ""
            if label == "ERROR" and shape is not None:
                shape_text = str(shape)
                if "google_" in shape_text or "/google/" in shape_text:
                    google_required_suffix = " Required by Google."
            if msg.startswith("Less than 1 values") and path:
                prefix = "Missing recommended property" if label == "WARN" else "Missing required property"
                lines.append((label, f"- {label} {path}: {prefix}.{google_required_suffix}{shape_str}"))
            elif "Value does not conform to Shape" in msg:
                fallback = "Value does not satisfy the required shape."
                if shape:
                    shape_msg = report_graph.value(shape, _SH.message)
                    if shape_msg:
                        fallback = str(shape_msg)
                    else:
                        hint = _or_shape_hint(report_graph, shape)
                        if hint:
                            fallback = hint
                if path:
                    lines.append((label, f"- {label} {path}: {fallback}{value_str}{shape_str}"))
                else:
                    lines.append((label, f"- {label} {fallback}{value_str}{focus_str}{shape_str}"))
            elif "must conform to one or more shapes" in msg:
                fallback = "Failed to satisfy one of the alternative constraints."
                if shape:
                    shape_msg = report_graph.value(shape, _SH.message)
                    if shape_msg:
                        fallback = str(shape_msg)
                    else:
                        hint = _or_shape_hint(report_graph, shape)
                        if hint:
                            fallback = hint
                if path:
                    lines.append((label, f"- {label} {path}: {fallback}{value_str}{shape_str}"))
                else:
                    lines.append((label, f"- {label} {fallback}{value_str}{focus_str}{shape_str}"))
            elif path and msg:
                lines.append((label, f"- {label} {path}: {msg}{value_str}{shape_str}"))
            elif msg:
                lines.append((label, f"- {label} {msg}{value_str}{focus_str}{shape_str}"))
            else:
                lines.append((label, f"- {label}"))
        lines.append(("text", ""))

    return header, lines


def _emit_validation_result(
    result,
    format: str,
    color: bool,
    report_file: str | None,
) -> None:
    if format not in {"pretty", "raw"}:
        raise UsageError("Invalid format. Use 'pretty' or 'raw'.")

    if format == "raw":
        output_text = result.report_text
        if report_file:
            Path(report_file).write_text(output_text, encoding="utf-8")
        else:
            typer.echo(output_text)
    else:
        header, lines = _format_pretty_lines(
            result.report_graph,
            result.data_graph,
            result.shape_source_map,
        )
        if report_file:
            output_text = "\n".join([header] + [line for _, line in lines]).rstrip() + "\n"
            Path(report_file).write_text(output_text, encoding="utf-8")
        else:
            status_color = {
                "Status: ERROR": typer.colors.RED,
                "Status: WARNING": typer.colors.YELLOW,
                "Status: INFO": typer.colors.BLUE,
                "Status: OK": typer.colors.GREEN,
            }
            header_color = None
            for prefix, fg in status_color.items():
                if header.startswith(prefix):
                    header_color = fg
                    break
            typer.secho(header, fg=header_color if color else None)
            for label, line in lines:
                if not line:
                    typer.echo("")
                    continue
                fg = None
                if color:
                    if label == "ERROR":
                        fg = typer.colors.RED
                    elif label == "WARN":
                        fg = typer.colors.YELLOW
                    elif label == "INFO":
                        fg = typer.colors.BLUE
                typer.secho(line, fg=fg)

    if not result.conforms:
        raise UsageError("SHACL validation failed.")
    if result.warning_count:
        raise UsageError("SHACL validation warnings found.")

    typer.echo("OK: no errors or warnings.")


def _warn_deprecated(command_hint: str) -> None:
    typer.secho(
        "Warning: `worai validate` is deprecated. "
        f"Use `{command_hint}` instead.",
        err=True,
        fg=typer.colors.YELLOW,
    )


def _is_url(value: str) -> bool:
    return value.startswith("http://") or value.startswith("https://")


class _ValidateGroup(TyperGroup):
    def resolve_command(self, ctx, args):
        try:
            return super().resolve_command(ctx, args)
        except click.UsageError:
            if not args:
                raise
            args.insert(0, "legacy")
            return super().resolve_command(ctx, args)


app = typer.Typer(add_completion=False, no_args_is_help=True, cls=_ValidateGroup)


def _run_jsonld_validation(
    input_file: str,
    shape: list[str] | None,
    report_file: str | None,
    format: str,
    color: bool,
) -> None:
    result = _validate_jsonld_input(input_file, shape)
    _emit_validation_result(result, format=format, color=color, report_file=report_file)


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    list_shapes: bool = typer.Option(
        False,
        "--list-shapes",
        help="List available packaged shapes and exit.",
    ),
) -> None:
    if list_shapes:
        _list_shapes()
        raise typer.Exit()
    if ctx.invoked_subcommand is None:
        raise UsageError("Missing input file or URL.")


@app.command(
    "jsonld",
    no_args_is_help=True,
    help="Deprecated. Use `worai graph validate` for RDF files/URLs.",
)
def jsonld(
    input_file: str = typer.Argument(
        ...,
        help="Path or URL to JSON-LD input.",
    ),
    shape: list[str] | None = typer.Option(
        None,
        "--shape",
        "-s",
        help="Shape path or packaged shape name (default: all packaged shapes).",
        show_default=False,
    ),
    report_file: str | None = typer.Option(
        None,
        "--report-file",
        help="Write the SHACL report text to this file.",
    ),
    format: str = typer.Option(
        "pretty",
        "--format",
        help="Output format: pretty or raw.",
        show_default=True,
    ),
    color: bool = typer.Option(
        True,
        "--color/--no-color",
        help="Colorize pretty output.",
        show_default=True,
    ),
) -> None:
    _warn_deprecated("worai graph validate <file-or-url> [<file-or-url> ...]")
    _run_jsonld_validation(input_file, shape, report_file, format, color)


def _list_shapes() -> None:
    from wordlift_sdk.validation import list_shape_names

    for name in list_shape_names():
        typer.echo(name)


def _validate_jsonld_input(input_value: str, shape: list[str] | None):
    try:
        return validate_file(input_value, shape)
    except Exception as exc:  # pragma: no cover - depends on SDK error detail
        if _is_url(input_value):
            raise UsageError(
                "Failed to validate URL as JSON-LD. "
                "If this is a webpage URL, use `worai structured-data validate page <url>`."
            ) from exc
        raise


@app.command("legacy", hidden=True, no_args_is_help=True)
def legacy(
    input_file: str = typer.Argument(
        ...,
        help="Deprecated. Use `worai validate jsonld` instead.",
    ),
    shape: list[str] | None = typer.Option(
        None,
        "--shape",
        "-s",
        help="Shape path or packaged shape name (default: all packaged shapes).",
        show_default=False,
    ),
    report_file: str | None = typer.Option(
        None,
        "--report-file",
        help="Write the SHACL report text to this file.",
    ),
    format: str = typer.Option(
        "pretty",
        "--format",
        help="Output format: pretty or raw.",
        show_default=True,
    ),
    color: bool = typer.Option(
        True,
        "--color/--no-color",
        help="Colorize pretty output.",
        show_default=True,
    ),
) -> None:
    _warn_deprecated("worai graph validate <file-or-url> [<file-or-url> ...]")
    _run_jsonld_validation(input_file, shape, report_file, format, color)



from __future__ import annotations

from pathlib import Path

from rdflib import Graph, URIRef

from worai.errors import UsageError

from .models import UrlRecord

SCHEMA_URL = URIRef("http://schema.org/url")
SEOVOC_HTML = URIRef("https://w3id.org/seovoc/html")


def _iter_ttl_files(path: Path) -> list[Path]:
    if path.is_file():
        if path.suffix.lower() != ".ttl":
            raise UsageError(f"Debug-cloud source file must be .ttl: {path}")
        return [path]
    if not path.exists():
        raise UsageError(f"Debug-cloud source path does not exist: {path}")
    if not path.is_dir():
        raise UsageError(f"Debug-cloud source must be a .ttl file or a directory: {path}")
    return sorted(item for item in path.rglob("*.ttl") if item.is_file())


def _first_entity_record(ttl_path: Path) -> UrlRecord | None:
    graph = Graph()
    graph.parse(ttl_path, format="turtle")

    fallback: UrlRecord | None = None
    for subject in graph.subjects(unique=True):
        url_value = graph.value(subject=subject, predicate=SCHEMA_URL)
        html_value = graph.value(subject=subject, predicate=SEOVOC_HTML)
        if url_value is None:
            continue
        url_text = str(url_value).strip()
        html_text = str(html_value) if html_value is not None else None
        if not url_text:
            continue
        record = UrlRecord(
            url=url_text,
            html=html_text,
            source_type="debug-cloud",
            source_ref=str(ttl_path),
            metadata={"subject": str(subject)},
        )
        # Prefer entities that include cached HTML. Some debug-cloud files
        # contain duplicate schema:url subjects where only one has seovoc:html.
        if html_text:
            return record
        if fallback is None:
            fallback = record

    return fallback


def load_records_from_debug_cloud(source: str) -> list[UrlRecord]:
    records: list[UrlRecord] = []
    for ttl_file in _iter_ttl_files(Path(source).expanduser()):
        record = _first_entity_record(ttl_file)
        if record is not None:
            records.append(record)
    return records



from __future__ import annotations

from worai.core import wordlift


class _FakeResponse:
    status_code = 200

    @staticmethod
    def json():
        return {"data": {"ok": True}}


def test_graphql_request_sets_private_header_when_enabled(monkeypatch) -> None:
    seen: dict[str, str] = {}

    def _stub_post(_endpoint, headers=None, data=None, timeout=60):
        seen.update(headers or {})
        return _FakeResponse()

    monkeypatch.setattr(wordlift.requests, "post", _stub_post)
    wordlift.graphql_request(
        "https://api.wordlift.io/graphql",
        "wl_123",
        "query { entities { iri } }",
        include_private=True,
    )

    assert seen["X-include-Private"] == "true"


def test_graphql_request_skips_private_header_by_default(monkeypatch) -> None:
    seen: dict[str, str] = {}

    def _stub_post(_endpoint, headers=None, data=None, timeout=60):
        seen.update(headers or {})
        return _FakeResponse()

    monkeypatch.setattr(wordlift.requests, "post", _stub_post)
    wordlift.graphql_request(
        "https://api.wordlift.io/graphql",
        "wl_123",
        "query { entities { iri } }",
    )

    assert "X-include-Private" not in seen


def test_resolve_api_key_uses_wordlift_api_key_env(monkeypatch) -> None:
    monkeypatch.setenv("WORDLIFT_API_KEY", "wl_alt")

    assert wordlift.resolve_api_key() == "wl_alt"


def test_resolve_api_key_ignores_wordlift_key_env(monkeypatch) -> None:
    monkeypatch.setenv("WORDLIFT_KEY", "wl_primary")
    monkeypatch.setenv("WORDLIFT_API_KEY", "wl_alt")

    assert wordlift.resolve_api_key() == "wl_alt"



from __future__ import annotations

from pathlib import Path

from worai.config import CONFIG_FILENAMES, load_config, resolve_config_path


def test_load_config_profile(tmp_path: Path, monkeypatch) -> None:
    config_path = tmp_path / "worai.toml"
    config_path.write_text(
        """
        [defaults]
        timeout = 10

        [profile.dev]
        timeout = 20
        """
    )

    cfg = load_config(str(config_path), "dev", env={})
    assert cfg.get("defaults.timeout") == 10
    assert cfg.get("timeout") == 20


def test_env_config_override(tmp_path: Path, monkeypatch) -> None:
    config_path = tmp_path / "worai.toml"
    config_path.write_text("[defaults]\nlog_level = 'info'\n")

    env = {"WORAI_CONFIG": str(config_path)}
    cfg = load_config(None, None, env=env)
    assert cfg.get("defaults.log_level") == "info"


def test_resolve_config_path_prefers_cli_option(tmp_path: Path) -> None:
    config_path = tmp_path / "myconfig.toml"
    resolved = resolve_config_path(str(config_path), env={})
    assert resolved == config_path.resolve()


def test_resolve_config_path_uses_env(tmp_path: Path) -> None:
    config_path = tmp_path / "envconfig.toml"
    env = {"WORAI_CONFIG": str(config_path)}
    resolved = resolve_config_path(None, env=env)
    assert resolved == config_path.resolve()


def test_resolve_config_path_defaults_when_no_files(monkeypatch) -> None:
    monkeypatch.setattr(Path, "exists", lambda _self: False)
    resolved = resolve_config_path(None, env={})
    assert resolved == Path(CONFIG_FILENAMES[0]).expanduser().resolve()


def test_resolve_config_path_uses_nearest_parent_worai_toml(tmp_path: Path, monkeypatch) -> None:
    project = tmp_path / "project"
    nested = project / "a" / "b"
    nested.mkdir(parents=True)
    config_path = project / "worai.toml"
    config_path.write_text("[profiles.default]\napi_key='wl'\n", encoding="utf-8")

    monkeypatch.chdir(nested)
    resolved = resolve_config_path(None, env={})
    assert resolved == config_path.resolve()



"""CLI wrapper for canonicalize-duplicate-pages."""

from __future__ import annotations

import typer

from worai.core.canonicalize_duplicate_pages import CanonicalizeOptions, run as canonicalize_run
from worai.core.profile_loader import resolve_profile_api_key
from worai.core.wordlift import DEFAULT_GRAPHQL_ENDPOINT
from worai.errors import UsageError

app = typer.Typer(add_completion=False, no_args_is_help=True, invoke_without_command=True)


@app.callback(invoke_without_command=True)
def run(
    ctx: typer.Context,
    input_csv: str = typer.Option(..., "--input", help="Input GSC CSV from worai google-search-console."),
    output_csv: str = typer.Option("canonical_targets.csv", "--output", help="Output CSV path."),
    url_regex: str | None = typer.Option(None, "--url-regex", help="Regex to filter URLs of interest."),
    entity_type: str | None = typer.Option(
        None,
        "--entity-type",
        help="Expected schema type when multiple entities share the same URL (e.g., Product).",
    ),
    endpoint: str = typer.Option(DEFAULT_GRAPHQL_ENDPOINT, "--endpoint", help="WordLift GraphQL endpoint."),
    batch_size: int = typer.Option(25, "--batch-size", help="Batch size for GraphQL queries."),
    kpi_window: str = typer.Option("28d", "--kpi-window", help="KPI time window (7d, 28d, 3m)."),
    kpi_metric: str = typer.Option("clicks", "--kpi-metric", help="KPI metric (clicks, impressions, ctr)."),
) -> None:
    try:
        api_key, _, _ = resolve_profile_api_key(ctx)
    except ValueError as exc:
        raise UsageError(str(exc)) from exc
    if not api_key:
        raise UsageError("WORDLIFT_API_KEY is required (or set profiles.<name>.api_key in config).")

    options = CanonicalizeOptions(
        api_key=api_key,
        input_csv=input_csv,
        output_csv=output_csv,
        url_regex=url_regex,
        entity_type=entity_type,
        endpoint=endpoint,
        batch_size=batch_size,
        kpi_window=kpi_window,
        kpi_metric=kpi_metric,
    )
    canonicalize_run(options)



from __future__ import annotations

import csv
import json
from pathlib import Path
import sys
import types

import pytest
from typer.testing import CliRunner

from worai.commands import structured_data as structured_data_cmd
from worai.core.structured_data_inventory import (
    InventoryOptions,
    InventoryRow,
    build_inventory,
    run_inventory,
)
from worai.errors import UsageError


def _load_json_from_cli_output(output: str) -> dict[str, object]:
    start = output.find("{")
    if start < 0:
        raise AssertionError(f"No JSON payload found in output: {output!r}")
    return json.loads(output[start:])


def _install_fake_sdk_ingestion(monkeypatch: pytest.MonkeyPatch, callback) -> None:
    pkg = types.ModuleType("wordlift_sdk")
    mod = types.ModuleType("wordlift_sdk.ingestion")
    mod.create_structured_data_inventory_from_ingestion = callback
    pkg.ingestion = mod
    monkeypatch.setitem(sys.modules, "wordlift_sdk", pkg)
    monkeypatch.setitem(sys.modules, "wordlift_sdk.ingestion", mod)


def test_build_inventory_calls_sdk_with_urls_bundle(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    seen: dict[str, object] = {}

    class _FakeDataFrame:
        def to_dict(self, orient: str = "records"):
            assert orient == "records"
            return [
                {
                    "url": "https://example.com/a",
                    "faq_markup": "yes",
                    "faq_markup_from_graph": "no",
                    "types": "WebPage",
                    "structured_data": "{}",
                }
            ]

    def _fake_create(**kwargs):
        seen["kwargs"] = kwargs
        return _FakeDataFrame()

    _install_fake_sdk_ingestion(monkeypatch, _fake_create)

    source = tmp_path / "urls.txt"
    source.write_text("https://example.com/a\nhttps://example.com/b\n", encoding="utf-8")

    rows = build_inventory(
        InventoryOptions(
            source=str(source),
            api_key="key",
            output=str(tmp_path / "out.csv"),
            ingest_source="urls",
            ingest_loader="playwright",
            ingest_passthrough_when_html=False,
            ingest_timeout_ms=10000,
            ingest_retry_attempts=1,
            ingest_retry_backoff_ms=0,
            playwright_wait_until="domcontentloaded",
            url_regex=r"/a$",
        )
    )

    assert len(rows) == 1
    assert rows[0].url == "https://example.com/a"
    kwargs = seen["kwargs"]
    assert kwargs["api_key"] == "key"
    assert kwargs["base_url"] == "https://api.wordlift.io"
    assert kwargs["output_csv"] == str(tmp_path / "out.csv")
    bundle = kwargs["source_bundle"]
    assert bundle["INGEST_SOURCE"] == "urls"
    assert bundle["INGEST_LOADER"] == "playwright"
    assert bundle["INGEST_PASSTHROUGH_WHEN_HTML"] is False
    assert bundle["INGEST_TIMEOUT_MS"] == 10000
    assert bundle["INGEST_RETRY_ATTEMPTS"] == 1
    assert bundle["INGEST_RETRY_BACKOFF_MS"] == 0
    assert bundle["PLAYWRIGHT_WAIT_UNTIL"] == "domcontentloaded"
    assert bundle["URL_REGEX"] == r"/a$"
    assert bundle["URLS"] == ["https://example.com/a", "https://example.com/b"]


def test_build_inventory_auto_infers_sitemap(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    seen: dict[str, object] = {}

    class _FakeDataFrame:
        def to_dict(self, orient: str = "records"):
            return []

    def _fake_create(**kwargs):
        seen["kwargs"] = kwargs
        return _FakeDataFrame()

    _install_fake_sdk_ingestion(monkeypatch, _fake_create)

    rows = build_inventory(
        InventoryOptions(
            source="https://example.com/sitemap.xml",
            api_key="key",
            output=str(tmp_path / "out.csv"),
        )
    )
    assert rows == []
    bundle = seen["kwargs"]["source_bundle"]
    assert bundle["INGEST_SOURCE"] == "sitemap"
    assert bundle["INGEST_LOADER"] == "web_scrape_api"
    assert "INGEST_TIMEOUT_MS" not in bundle
    assert "PLAYWRIGHT_WAIT_UNTIL" not in bundle
    assert bundle["SITEMAP_URL"] == "https://example.com/sitemap.xml"


def test_build_inventory_source_type_debug_cloud_maps_to_local(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    seen: dict[str, object] = {}

    class _FakeDataFrame:
        def to_dict(self, orient: str = "records"):
            return []

    def _fake_create(**kwargs):
        seen["kwargs"] = kwargs
        return _FakeDataFrame()

    _install_fake_sdk_ingestion(monkeypatch, _fake_create)

    source = tmp_path / "cloud"
    source.mkdir()

    build_inventory(
        InventoryOptions(
            source=str(source),
            source_type="debug-cloud",
            api_key="key",
            output=str(tmp_path / "out.csv"),
        )
    )

    bundle = seen["kwargs"]["source_bundle"]
    assert bundle["INGEST_SOURCE"] == "local"
    assert bundle["INGEST_LOCAL_ITEMS"] == []


def test_build_inventory_requires_sheet_name_for_sheets_source(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    _install_fake_sdk_ingestion(monkeypatch, lambda **_kwargs: None)

    with pytest.raises(UsageError, match="--sheet-name is required"):
        build_inventory(
            InventoryOptions(
                source="https://docs.google.com/spreadsheets/d/12345678901234567890/edit",
                api_key="key",
                output=str(tmp_path / "out.csv"),
                ingest_source="sheets",
            )
        )


def test_build_inventory_reads_csv_when_sdk_returns_none(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    def _fake_create(**kwargs):
        with open(kwargs["output_csv"], "w", encoding="utf-8", newline="") as handle:
            writer = csv.DictWriter(
                handle,
                fieldnames=["url", "faq_markup", "faq_markup_from_graph", "types", "structured_data"],
            )
            writer.writeheader()
            writer.writerow(
                {
                    "url": "https://example.com/a",
                    "faq_markup": "yes",
                    "faq_markup_from_graph": "yes",
                    "types": "WebPage",
                    "structured_data": "{}",
                }
            )
        return None

    _install_fake_sdk_ingestion(monkeypatch, _fake_create)

    rows = build_inventory(
        InventoryOptions(
            source="https://example.com/sitemap.xml",
            api_key="key",
            output=str(tmp_path / "out.csv"),
        )
    )

    assert len(rows) == 1
    assert rows[0].faq_markup == "yes"


def test_run_inventory_requires_exactly_one_destination() -> None:
    options = InventoryOptions(
        source="https://example.com/sitemap.xml",
        api_key="key",
    )
    with pytest.raises(UsageError):
        run_inventory(options)

    options = InventoryOptions(
        source="https://example.com/sitemap.xml",
        api_key="key",
        output="out.csv",
        destination_spreadsheet_id="spreadsheet-id",
    )
    with pytest.raises(UsageError):
        run_inventory(options)


def test_run_inventory_requires_destination_sheet_name() -> None:
    options = InventoryOptions(
        source="https://example.com/sitemap.xml",
        api_key="key",
        destination_spreadsheet_id="spreadsheet-id",
    )
    with pytest.raises(UsageError):
        run_inventory(options)


def test_run_inventory_writes_csv(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    rows = [
        InventoryRow(
            url="https://example.com/a",
            faq_markup="no",
            faq_markup_from_graph="no",
            types="WebPage",
            structured_data='{"@context":"https://schema.org","@graph":[]}',
        )
    ]

    monkeypatch.setattr(
        "worai.core.structured_data_inventory.build_inventory",
        lambda _options, on_progress=None: rows,
    )

    output_path = tmp_path / "inventory.csv"
    run_inventory(
        InventoryOptions(
            source="https://example.com/sitemap.xml",
            api_key="key",
            output=str(output_path),
        )
    )

    with output_path.open("r", encoding="utf-8") as handle:
        reader = csv.DictReader(handle)
        records = list(reader)

    assert len(records) == 1
    assert records[0]["url"] == "https://example.com/a"
    assert records[0]["types"] == "WebPage"


def test_inventory_command_builds_options_and_emits_summary(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options, on_progress=None):
        assert callable(on_progress)
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        ["inventory", "https://example.com/sitemap.xml", "--output", "report.csv"],
    )

    assert result.exit_code == 0
    payload = _load_json_from_cli_output(result.output)
    assert payload["total"] == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.source == "https://example.com/sitemap.xml"
    assert options.output == "report.csv"
    assert options.concurrency == "auto"


def test_inventory_command_sets_destination_sheet_options(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options, on_progress=None):
        assert callable(on_progress)
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        [
            "inventory",
            "https://example.com/sitemap.xml",
            "--destination-sheet-id",
            "spreadsheet-id",
            "--destination-sheet-name",
            "Inventory",
        ],
    )

    assert result.exit_code == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.destination_spreadsheet_id == "spreadsheet-id"
    assert options.destination_sheet_name == "Inventory"


def test_inventory_command_sets_concurrency(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options, on_progress=None):
        assert callable(on_progress)
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        [
            "inventory",
            "https://example.com/sitemap.xml",
            "--output",
            "report.csv",
            "--concurrency",
            "3",
        ],
    )

    assert result.exit_code == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.concurrency == "3"


def test_inventory_command_sets_source_type(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options, on_progress=None):
        assert callable(on_progress)
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        [
            "inventory",
            "/tmp/debug_cloud",
            "--output",
            "report.csv",
            "--source-type",
            "debug-cloud",
        ],
    )

    assert result.exit_code == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.source_type == "debug-cloud"


def test_inventory_command_sets_new_ingest_flags(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options, on_progress=None):
        assert callable(on_progress)
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        [
            "inventory",
            "https://example.com/sitemap.xml",
            "--output",
            "report.csv",
            "--ingest-source",
            "sitemap",
            "--ingest-loader",
            "web_scrape_api",
            "--url-regex",
            "/blog/",
            "--no-ingest-passthrough-when-html",
        ],
    )

    assert result.exit_code == 0
    payload = _load_json_from_cli_output(result.output)
    assert payload["url_regex"] == "/blog/"
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.ingest_source == "sitemap"
    assert options.ingest_loader == "web_scrape_api"
    assert options.url_regex == "/blog/"
    assert options.ingest_passthrough_when_html is False


def test_inventory_progress_callback_lifecycle_and_zero_case() -> None:
    bars: list[object] = []

    class _Bar:
        def __init__(self, *, total, desc, unit, leave):
            self.total = total
            self.desc = desc
            self.unit = unit
            self.leave = leave
            self.updates: list[int] = []
            self.closed = False

        def update(self, delta: int):
            self.updates.append(delta)

        def refresh(self):
            return None

        def close(self):
            self.closed = True

    def _factory(**kwargs):
        bar = _Bar(**kwargs)
        bars.append(bar)
        return bar

    callback, close = structured_data_cmd._build_inventory_progress_callback(
        enabled=True,
        progress_factory=_factory,
    )
    callback({"event": "inventory.progress.started", "meta": {"total": 3}})
    callback({"event": "inventory.progress.updated", "meta": {"completed": 1, "total": 3}})
    callback({"event": "inventory.progress.updated", "meta": {"completed": 1, "total": 3}})
    callback({"event": "inventory.progress.updated", "meta": {"completed": 3, "total": 3}})
    callback({"event": "inventory.progress.completed", "meta": {"completed": 3, "total": 3}})
    close()

    assert len(bars) == 1
    assert bars[0].updates == [1, 2]
    assert bars[0].closed is True

    zero_bars: list[object] = []

    def _zero_factory(**kwargs):
        bar = _Bar(**kwargs)
        zero_bars.append(bar)
        return bar

    callback, close = structured_data_cmd._build_inventory_progress_callback(
        enabled=True,
        progress_factory=_zero_factory,
    )
    callback({"event": "inventory.progress.started", "meta": {"total": 0}})
    callback({"event": "inventory.progress.updated", "meta": {"completed": 0, "total": 0}})
    callback({"event": "inventory.progress.completed", "meta": {"completed": 0, "total": 0}})
    close()

    assert len(zero_bars) == 1
    assert zero_bars[0].updates == []
    assert zero_bars[0].closed is True


def test_inventory_progress_callback_reports_events_when_progress_disabled() -> None:
    seen_events: list[dict[str, object]] = []

    callback, close = structured_data_cmd._build_inventory_progress_callback(
        enabled=False,
        progress_factory=None,
        event_reporter=lambda event: seen_events.append(event),
    )
    callback({"event": "inventory.progress.started", "meta": {"total": 2}})
    callback({"event": "inventory.progress.updated", "meta": {"completed": 1, "total": 2}})
    callback({"event": "inventory.progress.completed", "meta": {"completed": 2, "total": 2}})
    close()

    assert len(seen_events) == 3
    assert seen_events[0]["event"] == "inventory.progress.started"
    assert seen_events[-1]["event"] == "inventory.progress.completed"


def test_inventory_command_resolves_ingest_runtime_settings_from_profile(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options, on_progress=None):
        captured["options"] = options
        return []

    monkeypatch.setattr(
        structured_data_cmd,
        "load_profile_settings",
        lambda _ctx: (
            {
                "ingest.timeout_ms": "9000",
                "ingest.retry_attempts": "2",
                "ingest.retry_backoff_ms": "100",
                "ingest.playwright_wait_until": "domcontentloaded",
            },
            "default",
        ),
    )
    monkeypatch.setattr(
        structured_data_cmd,
        "resolve_profile_api_key",
        lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"),
    )
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    result = CliRunner().invoke(
        structured_data_cmd.app,
        ["inventory", "https://example.com/sitemap.xml", "--output", "report.csv"],
    )

    assert result.exit_code == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.ingest_timeout_ms == 9000
    assert options.ingest_retry_attempts == 2
    assert options.ingest_retry_backoff_ms == 100
    assert options.playwright_wait_until == "domcontentloaded"



"""wordlift-cli entry point — thin wrapper for `worai agent --agent-cli gemini`."""

from __future__ import annotations

import sys


def run() -> None:
    """Launch `worai agent --agent-cli gemini -- <extra-args>`."""
    extra = sys.argv[1:]
    sys.argv = [sys.argv[0], "agent", "--agent-cli", "gemini", "--", *extra]
    from worai.cli import run as _run

    _run()



"""Unit tests for worai.core.output."""

from __future__ import annotations

import csv
import io
import json
import sys
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from worai.core.output import OutputFormat, write_output

SAMPLE_ROWS = [
    {"name": "Alice", "score": 42},
    {"name": "Bob", "score": 7},
]


# ---------------------------------------------------------------------------
# table
# ---------------------------------------------------------------------------


def test_table_renders_without_error(capsys):
    write_output(SAMPLE_ROWS, OutputFormat.table)
    out = capsys.readouterr().out
    assert "Alice" in out
    assert "Bob" in out


def test_table_column_headers_present(capsys):
    write_output(SAMPLE_ROWS, OutputFormat.table)
    out = capsys.readouterr().out
    assert "name" in out
    assert "score" in out


def test_table_empty_rows_produces_no_output(capsys):
    write_output([], OutputFormat.table)
    out = capsys.readouterr().out
    assert out == ""


def test_table_writes_to_file(tmp_path):
    dest = tmp_path / "table.txt"
    write_output(SAMPLE_ROWS, OutputFormat.table, dest=dest)
    content = dest.read_text()
    assert "Alice" in content
    assert "Bob" in content


# ---------------------------------------------------------------------------
# json
# ---------------------------------------------------------------------------


def test_json_stdout_is_valid_array(capsys):
    write_output(SAMPLE_ROWS, OutputFormat.json)
    out = capsys.readouterr().out
    parsed = json.loads(out)
    assert parsed == SAMPLE_ROWS


def test_json_to_file(tmp_path):
    dest = tmp_path / "out.json"
    write_output(SAMPLE_ROWS, OutputFormat.json, dest=dest)
    parsed = json.loads(dest.read_text())
    assert parsed == SAMPLE_ROWS


# ---------------------------------------------------------------------------
# csv
# ---------------------------------------------------------------------------


def test_csv_has_header_and_rows(capsys):
    write_output(SAMPLE_ROWS, OutputFormat.csv)
    out = capsys.readouterr().out
    reader = csv.DictReader(io.StringIO(out))
    rows = list(reader)
    assert rows == [{"name": "Alice", "score": "42"}, {"name": "Bob", "score": "7"}]


def test_csv_to_file_header_and_rows(tmp_path):
    dest = tmp_path / "out.csv"
    write_output(SAMPLE_ROWS, OutputFormat.csv, dest=dest)
    reader = csv.DictReader(dest.open(newline=""))
    rows = list(reader)
    assert len(rows) == 2
    assert rows[0]["name"] == "Alice"
    assert rows[1]["score"] == "7"


def test_csv_empty_rows_writes_only_empty_header(capsys):
    write_output([], OutputFormat.csv)
    out = capsys.readouterr().out
    # No fieldnames → empty header line only
    assert out == "\r\n"


# ---------------------------------------------------------------------------
# tsv
# ---------------------------------------------------------------------------


def test_tsv_has_header_and_rows(capsys):
    write_output(SAMPLE_ROWS, OutputFormat.tsv)
    out = capsys.readouterr().out
    reader = csv.DictReader(io.StringIO(out), delimiter="\t")
    rows = list(reader)
    assert rows[0]["name"] == "Alice"
    assert rows[1]["score"] == "7"


def test_tsv_uses_tab_delimiter(capsys):
    write_output(SAMPLE_ROWS, OutputFormat.tsv)
    out = capsys.readouterr().out
    first_line = out.splitlines()[0]
    assert "\t" in first_line
    assert "," not in first_line


# ---------------------------------------------------------------------------
# parquet
# ---------------------------------------------------------------------------


def test_parquet_requires_dest():
    with pytest.raises(ValueError, match="destination file path"):
        write_output(SAMPLE_ROWS, OutputFormat.parquet)


def test_parquet_writes_readable_file(tmp_path):
    pyarrow = pytest.importorskip("pyarrow")
    import pyarrow.parquet as pq

    dest = tmp_path / "out.parquet"
    write_output(SAMPLE_ROWS, OutputFormat.parquet, dest=dest)

    table = pq.read_table(str(dest))
    assert table.num_rows == 2
    assert "name" in table.column_names
    assert "score" in table.column_names


def test_parquet_missing_extra_raises_import_error(tmp_path):
    dest = tmp_path / "out.parquet"
    with patch.dict(sys.modules, {"pyarrow": None, "pyarrow.parquet": None}):
        with pytest.raises(ImportError, match=r"worai\[parquet\]"):
            write_output(SAMPLE_ROWS, OutputFormat.parquet, dest=dest)


# ---------------------------------------------------------------------------
# gsheets
# ---------------------------------------------------------------------------


def test_gsheets_missing_credentials_raises():
    with pytest.raises(ValueError, match="credentials_file"):
        write_output(SAMPLE_ROWS, OutputFormat.gsheets, spreadsheet_id="x", sheet_name="s")


def test_gsheets_missing_spreadsheet_id_raises():
    with pytest.raises(ValueError, match="spreadsheet_id"):
        write_output(SAMPLE_ROWS, OutputFormat.gsheets, credentials_file="c.json", sheet_name="s")


def test_gsheets_missing_sheet_name_raises():
    with pytest.raises(ValueError, match="sheet_name"):
        write_output(SAMPLE_ROWS, OutputFormat.gsheets, credentials_file="c.json", spreadsheet_id="x")


def test_gsheets_clears_and_writes_header_plus_rows():
    mock_worksheet = MagicMock()
    mock_spreadsheet = MagicMock()
    mock_spreadsheet.worksheet.return_value = mock_worksheet

    mock_gc = MagicMock()
    mock_gc.open_by_key.return_value = mock_spreadsheet

    mock_gspread = MagicMock()
    mock_gspread.service_account.return_value = mock_gc

    with patch.dict(sys.modules, {"gspread": mock_gspread}):
        write_output(
            SAMPLE_ROWS,
            OutputFormat.gsheets,
            credentials_file="svc.json",
            spreadsheet_id="sheet_id",
            sheet_name="Tab1",
        )

    mock_worksheet.clear.assert_called_once()
    mock_worksheet.update.assert_called_once()

    written = mock_worksheet.update.call_args[0][0]
    assert written[0] == ["name", "score"]        # header row
    assert written[1] == ["Alice", 42]             # first data row
    assert written[2] == ["Bob", 7]                # second data row


def test_gsheets_missing_extra_raises_import_error():
    with patch.dict(sys.modules, {"gspread": None}):
        with pytest.raises(ImportError, match=r"worai\[gsheets\]"):
            write_output(
                SAMPLE_ROWS,
                OutputFormat.gsheets,
                credentials_file="c.json",
                spreadsheet_id="s",
                sheet_name="t",
            )



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import pytest

from worai.core import canonicals as mod
from worai.errors import UsageError


def test_run_dedupe_with_service_account(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    captured: dict[str, object] = {}

    def _sdk(**kwargs):
        captured.update(kwargs)
        return [1, 2]

    monkeypatch.setattr(mod, "_load_oauth_credentials", lambda *_a, **_k: (_ for _ in ()).throw(AssertionError()))
    monkeypatch.setattr(mod, "_load_service_account_credentials", lambda *_a, **_k: object())
    monkeypatch.setitem(
        __import__("sys").modules,
        "wordlift_sdk.google_search_console",
        SimpleNamespace(create_canonical_csv_from_gsc_impressions=_sdk),
    )

    out = mod.run_dedupe(
        mod.CanonicalsDedupeOptions(
            input_csv=str(tmp_path / "in.csv"),
            output_csv=str(tmp_path / "out.csv"),
            site_url="sc-domain:example.com",
            service_account="sa.json",
        )
    )
    assert out == [1, 2]
    assert "credentials" in captured


def test_load_oauth_credentials_uses_existing_token(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    token = tmp_path / "oauth.json"
    token.write_text("{}", encoding="utf-8")

    class _Creds:
        expired = False
        refresh_token = None
        valid = True

    monkeypatch.setitem(
        __import__("sys").modules,
        "wordlift_sdk.google_search_console",
        SimpleNamespace(load_authorized_user_credentials=lambda _p: _Creds()),
    )
    monkeypatch.setattr(mod, "load_credentials", lambda *_a, **_k: (_ for _ in ()).throw(AssertionError()))

    creds = mod._load_oauth_credentials(str(token), None, 8080)
    assert creds.valid is True


def test_load_oauth_credentials_requires_client_secrets_without_valid_token(tmp_path: Path) -> None:
    with pytest.raises(UsageError, match="Authentication is missing"):
        mod._load_oauth_credentials(str(tmp_path / "missing.json"), None, 8080)


def test_load_service_account_credentials_accepts_inline_json(monkeypatch: pytest.MonkeyPatch) -> None:
    seen = {}

    class _ServiceAccountCreds:
        @staticmethod
        def from_service_account_info(info, scopes):
            seen["info"] = info
            seen["scopes"] = scopes
            return object()

    monkeypatch.setattr(mod.service_account, "Credentials", _ServiceAccountCreds)
    creds = mod._load_service_account_credentials('{"client_email":"x@example.com"}')
    assert creds is not None
    assert seen["info"]["client_email"] == "x@example.com"


def test_load_service_account_credentials_rejects_invalid_value() -> None:
    with pytest.raises(UsageError, match="valid file path or a JSON body"):
        mod._load_service_account_credentials("not-a-path-and-not-json")



"""CLI wrapper for link-groups."""

from __future__ import annotations

import typer

from worai.core.link_groups import LinkGroupsOptions, run as link_groups_run
from worai.core.profile_loader import resolve_profile_api_key
from worai.errors import UsageError

app = typer.Typer(add_completion=False, no_args_is_help=True, invoke_without_command=True)


@app.callback(invoke_without_command=True)
def run(
    ctx: typer.Context,
    input_file: str = typer.Argument(..., help="Path to input CSV file."),
    output_format: str = typer.Option("turtle", "--format", "-f", help="RDF output format."),
    apply: bool = typer.Option(False, "--apply", help="Apply changes via WordLift API."),
    dry_run: bool = typer.Option(False, "--dry-run", help="Show curl commands without executing them."),
    concurrency: int = typer.Option(2, "--concurrency", help="Number of concurrent API requests."),
    retries: int = typer.Option(3, "--retries", help="Number of retries for failed requests."),
) -> None:
    api_key = None
    if apply:
        try:
            api_key, _, _ = resolve_profile_api_key(ctx)
        except ValueError as exc:
            raise UsageError(str(exc)) from exc
        if not api_key:
            raise UsageError("WORDLIFT_API_KEY is required when using --apply.")

    options = LinkGroupsOptions(
        input_file=input_file,
        output_format=output_format,
        apply=apply,
        dry_run=dry_run,
        api_key=api_key,
        concurrency=concurrency,
        retries=retries,
    )

    output = link_groups_run(options)
    if output:
        typer.echo(output)



from __future__ import annotations

from pathlib import Path

import pytest
import requests

from worai.core import find_faq_page_wrong_type as mod


def test_payload_and_curl_generation() -> None:
    payload = mod._build_patch_payload("https://example.com/e", replace_type=False)
    assert payload[0]["op"] == "add"

    payload_replace = mod._build_patch_payload("https://example.com/e", replace_type=True)
    assert payload_replace[0]["op"] == "remove"
    assert payload_replace[1]["op"] == "add"

    curl = mod.generate_curl_command("https://example.com/e", replace_type=True)
    assert "curl -L -X PATCH" in curl
    assert "remove" in curl


def test_patch_entity_success_and_retries(monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]) -> None:
    calls: list[dict] = []

    class _Resp:
        status_code = 200

        @staticmethod
        def raise_for_status() -> None:
            return None

    def _ok_patch(url: str, headers: dict, json: list[dict], timeout: int):
        calls.append({"url": url, "headers": headers, "json": json, "timeout": timeout})
        return _Resp()

    monkeypatch.setattr(mod.requests, "patch", _ok_patch)
    mod.patch_entity("https://example.com/e", "wl", replace_type=True)
    assert calls
    assert calls[0]["timeout"] == 15
    assert calls[0]["headers"]["Authorization"] == "Key wl"

    def _fail_patch(*_a, **_k):
        raise requests.exceptions.RequestException("x")

    monkeypatch.setattr(mod.requests, "patch", _fail_patch)
    monkeypatch.setattr(mod.time, "sleep", lambda _s: None)
    mod.patch_entity("https://example.com/e", "wl", replace_type=False)
    out = capsys.readouterr().out
    assert "Giving up" in out


def test_find_entities_and_run_modes(monkeypatch: pytest.MonkeyPatch, tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    ttl = tmp_path / "data.ttl"
    ttl.write_text(
        """
        @prefix schema: <http://schema.org/> .
        @prefix ex: <https://example.com/> .
        ex:e1 schema:mainEntity ex:o .
        """,
        encoding="utf-8",
    )
    entities = mod.find_entities(str(ttl))
    assert entities == ["https://example.com/e1"]

    monkeypatch.setattr(mod, "find_entities", lambda _p: [])
    assert mod.run(mod.FaqFixOptions(file_path="x.ttl", mode="find")) == []

    monkeypatch.setattr(mod, "find_entities", lambda _p: ["https://example.com/e1"])
    assert mod.run(mod.FaqFixOptions(file_path="x.ttl", mode="find")) == ["https://example.com/e1"]

    mod.run(mod.FaqFixOptions(file_path="x.ttl", mode="dry-run", replace_type=True))
    assert "Sample cURL command" in capsys.readouterr().out

    with pytest.raises(RuntimeError, match="API key required"):
        mod.run(mod.FaqFixOptions(file_path="x.ttl", mode="patch"))

    seen: dict[str, object] = {}
    monkeypatch.setattr(
        mod,
        "patch_entity",
        lambda entity_id, api_key, replace_type=False: seen.update(
            {"entity_id": entity_id, "api_key": api_key, "replace_type": replace_type}
        ),
    )
    mod.run(mod.FaqFixOptions(file_path="x.ttl", mode="patch", api_key="wl", replace_type=True))
    assert seen["api_key"] == "wl"
    assert seen["replace_type"] is True



"""Agent-backed structured data generation."""

from __future__ import annotations

from typing import Callable
from pathlib import Path

from wordlift_sdk.structured_data import AgentGenerator as EngineAgentGenerator


class AgentGenerator:
    """Generates YARRRML and JSON-LD via the agent."""

    def __init__(self, engine: EngineAgentGenerator | None = None) -> None:
        self._engine = engine or EngineAgentGenerator()

    def generate(
        self,
        url: str,
        html: str,
        xhtml: str,
        cleaned_xhtml: str,
        api_key: str,
        dataset_uri: str,
        target_type: str,
        workdir: Path,
        debug: bool,
        max_retries: int,
        max_nesting_depth: int,
        quality_check: bool,
        log: Callable[[str], None],
    ) -> tuple[str, dict]:
        return self._engine.generate_from_agent(
            url,
            html,
            xhtml,
            cleaned_xhtml,
            api_key,
            dataset_uri,
            target_type,
            workdir,
            debug=debug,
            max_retries=max_retries,
            max_nesting_depth=max_nesting_depth,
            quality_check=quality_check,
            log=log,
        )



"""Tests for worai.core.entity_matrix and the entity-matrix CLI command."""

from __future__ import annotations

import textwrap

import pytest
from typer.testing import CliRunner

from worai.cli import app
from worai.core.entity_matrix import EntityMatrixOptions, build_entity_matrix

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

_TURTLE_SIMPLE = textwrap.dedent("""\
    @prefix schema: <http://schema.org/> .

    <https://example.com/article>
        a schema:Article ;
        schema:url "https://example.com/article" ;
        schema:about <https://example.com/entity/topic> .

    <https://example.com/article/faq>
        a schema:FAQPage .

    <https://example.com/article/faq/q1>
        a schema:Question .

    <https://example.com/entity/topic>
        a schema:Thing .
""")

_TURTLE_BLOG = textwrap.dedent("""\
    @prefix schema: <http://schema.org/> .

    <https://example.com/blog/post-1>
        a schema:Article ;
        schema:url "https://example.com/blog/post-1" .

    <https://example.com/blog/post-2>
        a schema:Article ;
        schema:url "https://example.com/blog/post-2" .

    <https://example.com/blog/post-3>
        a schema:Article, schema:BlogPosting ;
        schema:url "https://example.com/blog/post-3" .
""")

# Two distinct signature groups under the same parent (/mixed/*), both with ≥2
# URLs → the SDK emits two separate "parent/*" rows; the CLI must merge them.
_TURTLE_MIXED = textwrap.dedent("""\
    @prefix schema: <http://schema.org/> .

    <https://example.com/mixed/a1>
        a schema:Article ;
        schema:url "https://example.com/mixed/a1" .

    <https://example.com/mixed/a2>
        a schema:Article ;
        schema:url "https://example.com/mixed/a2" .

    <https://example.com/mixed/b1>
        a schema:BlogPosting ;
        schema:url "https://example.com/mixed/b1" .

    <https://example.com/mixed/b2>
        a schema:BlogPosting ;
        schema:url "https://example.com/mixed/b2" .
""")


def _write_ttl(tmp_path, content: str, name: str = "graph.ttl"):
    path = tmp_path / name
    path.write_text(content, encoding="utf-8")
    return str(path)


# ---------------------------------------------------------------------------
# Unit: subgraph coverage (tested via build_entity_matrix)
# ---------------------------------------------------------------------------


def test_subgraph_includes_iri_prefix_children(tmp_path):
    # FAQPage and Question are IRI-prefix children of the article → included
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path))
    article_row = next(r for r in rows if r["url"] == "https://example.com/article")
    assert article_row.get("FAQPage", 0) >= 1
    assert article_row.get("Question", 0) >= 1


def test_subgraph_includes_referenced_entity(tmp_path):
    # ex:entity/topic (schema:Thing) is linked via schema:about → included
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path))
    article_row = next(r for r in rows if r["url"] == "https://example.com/article")
    assert article_row.get("Thing", 0) >= 1


# ---------------------------------------------------------------------------
# Unit: build_entity_matrix
# ---------------------------------------------------------------------------


def test_matrix_has_url_column(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path))
    assert all("url" in row for row in rows)


def test_matrix_article_row_has_faqpage(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path))
    article_row = next(r for r in rows if r["url"] == "https://example.com/article")
    # FAQPage is a child of the article → included in its subgraph
    assert article_row.get("FAQPage", 0) >= 1


def test_matrix_article_row_has_referenced_thing(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path))
    article_row = next(r for r in rows if r["url"] == "https://example.com/article")
    # ex:topic (schema:Thing) is directly referenced → included
    assert article_row.get("Thing", 0) >= 1


def test_matrix_exclude_type(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    rows = build_entity_matrix(
        EntityMatrixOptions(filename=path, exclude_types=["FAQPage"])
    )
    assert all("FAQPage" not in row for row in rows)


def test_matrix_file_not_found():
    with pytest.raises(FileNotFoundError):
        build_entity_matrix(EntityMatrixOptions(filename="/no/such/file.ttl"))


def test_matrix_counts_are_integers(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path))
    for row in rows:
        for k, v in row.items():
            if k != "url":
                assert isinstance(v, int)


# ---------------------------------------------------------------------------
# Unit: clustering
# ---------------------------------------------------------------------------


def test_cluster_collapses_identical_type_signature(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_BLOG)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path, cluster=True))
    urls = [r["url"] for r in rows]
    # post-1 and post-2 share identical types → collapsed into /blog/*
    assert "https://example.com/blog/*" in urls
    # post-3 has an extra BlogPosting → kept separate
    assert "https://example.com/blog/post-3" in urls
    # total rows: one wildcard + one individual
    assert len(urls) == 2


def test_cluster_keeps_different_signatures_separate(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_BLOG)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path, cluster=True))
    urls = [r["url"] for r in rows]
    # post-3 has BlogPosting in addition to Article → kept separate
    assert "https://example.com/blog/post-3" in urls


def test_cluster_sums_counts(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_BLOG)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path, cluster=True))
    wildcard = next((r for r in rows if r["url"].endswith("/*")), None)
    assert wildcard is not None
    # post-1 and post-2 each have 1 Article → sum = 2
    assert wildcard.get("Article", 0) == 2


def test_cluster_no_duplicate_wildcard_urls(tmp_path):
    # Two distinct signature sub-groups under /mixed/, each with 2 URLs.
    # The SDK emits one wildcard row per sub-group; the CLI must merge them
    # so there is exactly one "https://example.com/mixed/*" row.
    path = _write_ttl(tmp_path, _TURTLE_MIXED)
    rows = build_entity_matrix(EntityMatrixOptions(filename=path, cluster=True))
    wildcard_urls = [r["url"] for r in rows if r["url"].endswith("/*")]
    assert wildcard_urls.count("https://example.com/mixed/*") == 1
    # Merged row must carry both Article and BlogPosting counts
    merged = next(r for r in rows if r["url"] == "https://example.com/mixed/*")
    assert merged.get("Article", 0) == 2
    assert merged.get("BlogPosting", 0) == 2


# ---------------------------------------------------------------------------
# CLI integration
# ---------------------------------------------------------------------------

_runner = CliRunner()


def test_cli_table_output(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    result = _runner.invoke(app, ["entity-matrix", str(path)])
    assert result.exit_code == 0
    assert "url" in result.output


def test_cli_csv_to_file(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    out = tmp_path / "out.csv"
    result = _runner.invoke(
        app, ["entity-matrix", str(path), "--format", "csv", "--output", str(out)]
    )
    assert result.exit_code == 0
    content = out.read_text()
    assert "url" in content
    assert "Article" in content


def test_cli_json_output(tmp_path):
    import json

    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    result = _runner.invoke(app, ["entity-matrix", str(path), "--format", "json"])
    assert result.exit_code == 0
    data = json.loads(result.output)
    assert isinstance(data, list)
    assert all("url" in row for row in data)


def test_cli_default_presence_markers(tmp_path):
    import json

    # _TURTLE_BLOG: post-1/post-2 have Article only; post-3 has Article+BlogPosting.
    # The BlogPosting column is absent (0) for post-1 → should render as "".
    path = _write_ttl(tmp_path, _TURTLE_BLOG)
    result = _runner.invoke(app, ["entity-matrix", str(path), "--format", "json"])
    assert result.exit_code == 0
    data = json.loads(result.output)
    post1 = next(r for r in data if r["url"] == "https://example.com/blog/post-1")
    post3 = next(r for r in data if r["url"] == "https://example.com/blog/post-3")
    # Present types → ✅
    assert post1.get("Article") == "✅"
    assert post3.get("BlogPosting") == "✅"
    # Absent type → ""
    assert post1.get("BlogPosting") == ""


def test_cli_count_flag_shows_integers(tmp_path):
    import json

    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    result = _runner.invoke(
        app, ["entity-matrix", str(path), "--format", "json", "--count"]
    )
    assert result.exit_code == 0
    data = json.loads(result.output)
    for row in data:
        for k, v in row.items():
            if k != "url":
                assert isinstance(v, int)


def test_cli_exclude_type(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_SIMPLE)
    result = _runner.invoke(
        app, ["entity-matrix", str(path), "--format", "json", "--exclude-type", "FAQPage"]
    )
    import json

    assert result.exit_code == 0
    data = json.loads(result.output)
    assert all("FAQPage" not in row for row in data)


def test_cli_cluster_flag(tmp_path):
    path = _write_ttl(tmp_path, _TURTLE_BLOG)
    result = _runner.invoke(
        app, ["entity-matrix", str(path), "--format", "json", "--cluster"]
    )
    import json

    assert result.exit_code == 0
    data = json.loads(result.output)
    assert any(r["url"].endswith("/*") for r in data)


def test_cli_missing_file():
    result = _runner.invoke(app, ["entity-matrix", "/no/such/file.ttl"])
    assert result.exit_code != 0



"""WordLift API helpers."""

from __future__ import annotations

import json
import os
from dataclasses import dataclass
from typing import Iterable
from urllib.parse import quote_plus

import requests

DEFAULT_GRAPHQL_ENDPOINT = "https://api.wordlift.io/graphql"
DEFAULT_ENTITIES_ENDPOINT = "https://api.wordlift.io/entities"


@dataclass
class WordLiftConfig:
    api_key: str
    graphql_endpoint: str = DEFAULT_GRAPHQL_ENDPOINT
    entities_endpoint: str = DEFAULT_ENTITIES_ENDPOINT


def resolve_api_key(env_key: str = "WORDLIFT_API_KEY") -> str | None:
    return os.environ.get(env_key)


def graphql_request(
    endpoint: str,
    api_key: str,
    query: str,
    timeout: int = 60,
    include_private: bool = False,
) -> dict:
    headers = {
        "Authorization": f"Key {api_key}",
        "Content-Type": "application/json",
        "Accept": "application/json",
    }
    if include_private:
        headers["X-include-Private"] = "true"
    payload = {"query": query}
    resp = requests.post(endpoint, headers=headers, data=json.dumps(payload), timeout=timeout)
    if resp.status_code != 200:
        raise RuntimeError(f"GraphQL HTTP {resp.status_code}: {resp.text[:500]}")
    try:
        data = resp.json()
    except ValueError as exc:
        raise RuntimeError("Failed to decode GraphQL JSON response") from exc
    if data.get("errors"):
        raise RuntimeError(f"GraphQL errors: {data['errors']}")
    return data


def delete_entity(iri: str, api_key: str, include_children: bool = True, timeout: int = 60) -> bool:
    params = f"id={quote_plus(iri)}"
    if include_children:
        params = f"{params}&include_children=true"
    url = f"{DEFAULT_ENTITIES_ENDPOINT}?{params}"
    headers = {"Authorization": f"Key {api_key}"}
    resp = requests.delete(url, headers=headers, timeout=timeout)
    return 200 <= resp.status_code < 300


def delete_entities(
    iris: Iterable[str],
    api_key: str,
    include_children: bool = True,
    timeout: int = 60,
) -> dict[str, bool]:
    results: dict[str, bool] = {}
    for iri in iris:
        results[iri] = delete_entity(iri, api_key, include_children=include_children, timeout=timeout)
    return results
